#!/usr/bin/env python3

# Apply a Machine Learning algorithm to obtain predictions
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Ex: python 4_predict.py --directory dataset_test.csv --model model_cnn_classifier_2021_04_15-10_23_13


import os
import tqdm
import json
import logging
import tempfile
import argparse
import numpy as np
import pandas as pd
from PIL import Image
from typing import List
from pathlib import Path
from datetime import datetime

from {{package_name}} import utils
from {{package_name}}.preprocessing import preprocess
from {{package_name}}.models_training import utils_models

# Get logger
logger = logging.getLogger('{{package_name}}.3_predict')


def main(directory: str, model_dir: str, sep: str = '{{default_sep}}', encoding: str = '{{default_encoding}}') -> None:
    '''Gets a model's predictions on a given dataset

    Args:
        directory (str): Name of the test directory (actually a path relative to {{package_name}}-data)
            It must NOT be preprocessed
        model_dir (str): Name of the model to use (not a path, just the directory name)
    Kwargs:
        sep (str): Separator to use with the .csv files
        encoding (str): Encoding to use with the .csv files
    Raises:
        FileNotFoundError: If the directory does not exists in {{package_name}}-data
        NotADirectoryError: If the argument `directory` is not a directory
    '''

    ##############################################
    # Loading data
    ##############################################

    # Get data abs path
    data_path = utils.get_data_path()
    directory_path = os.path.join(data_path, directory)
    if not os.path.exists(directory_path):
        raise FileNotFoundError(f"{directory_path} path does not exist'")
    if not os.path.isdir(directory_path):
        raise NotADirectoryError(f"{directory_path} is not a valid directory")

    # Retrieve path/class/bboxes informations
    logger.info("Loading dataset ...")
    path_list, classes_or_bboxes_list, _, task_type = utils.read_folder(directory_path, sep=sep, encoding=encoding, accept_no_metadata=True)


    ##############################################
    # Load model
    ##############################################

    # Load model
    logger.info("Loading model ...")
    model, model_conf = utils_models.load_model(model_dir=model_dir)

    # Retrieve preprocessing
    preprocess_str = model_conf['preprocess_str']


    ##############################################
    # Preprocessing
    ##############################################

    # We'll create a temporary folder to save preprocessed images
    with tempfile.TemporaryDirectory(dir=utils.get_data_path()) as tmp_folder:
        # Preprocessing
        if preprocess_str != 'no_preprocess':
            new_path_list = apply_preprocessing(path_list, preprocess_str, directory_path, tmp_folder)
            df = pd.DataFrame({'file_path': new_path_list, 'original_file_path': path_list})
        else:
            logger.info("No preprocessing to be applied")
            df = pd.DataFrame({'file_path': path_list, 'original_file_path': path_list})


        ##############################################
        # Predictions
        ##############################################

        # Get predictions
        logger.info("Getting predictions ...")
        y_pred = model.predict(df, return_proba=False)
        df['predictions'] = list(model.inverse_transform(np.array(y_pred)))

    # Getting out of the context, all temporary data is deleted


    ##############################################
    # Save results
    ##############################################

    # Add filename & select final columns
    df['filename'] = [os.path.relpath(f, directory_path) for f in path_list]
    df = df[['original_file_path', 'filename', 'predictions']]

    # Save result
    logger.info("Saving results ...")
    save_dir = os.path.join(data_path, 'predictions', directory, datetime.now().strftime("predictions_%Y_%m_%d-%H_%M_%S"))
    if not os.path.isdir(save_dir):
        os.makedirs(save_dir)
    save_file = "predictions.csv"
    file_path = os.path.join(save_dir, save_file)
    df.to_csv(file_path, sep='{{default_sep}}', encoding='{{default_encoding}}', index=None)

    # Also save some info into a configs file
    conf_file = 'configurations.json'
    conf_path = os.path.join(save_dir, conf_file)
    conf = {
        'model_dir': model_dir,
        'preprocess_str': model_conf['preprocess_str'],
        'model_name': model_conf['model_name']
    }
    with open(conf_path, 'w', encoding='{{default_encoding}}') as f:
        json.dump(conf, f, indent=4)


    ##############################################
    # Get metrics
    ##############################################

    # Get metrics if classes_or_bboxes_list is not None
    if classes_or_bboxes_list is not None:
        # Change model directory to save dir & get preds
        model.model_dir = save_dir
        model.get_and_save_metrics(classes_or_bboxes_list, y_pred, list_files_x=list(df['original_file_path'].values), type_data='test')


def apply_preprocessing(path_list: List[str], preprocess_str: str, old_directory: str, tmp_folder: str) -> List[str]:
    '''Function to apply a preprocessing on an image directory

    Args:
        path_list (list<str>): path list of all images to be preprocessed
        preprocess_str (str): preprocessing to be applied
        old_directory (str): directory with the original images - path
        tmp_folder (str): directory where to save the preprocessed images - path
    Returns:
        list<str>: new paths list for the preprocessed images
    '''
    logger.info(f"Applying preprocessing {preprocess_str} on test images ...")

    # Get preprocessor
    preprocessor = preprocess.get_preprocessor(preprocess_str)

    # Process by chunks of 100s to avoid memory issues
    new_path_list = []
    chunks_limits = utils.get_chunk_limits(pd.Series(path_list), chunksize=100)
    for limits in tqdm.tqdm(chunks_limits):
        # Manage data
        min_l, max_l = limits[0], limits[1]
        tmp_path_list = path_list[min_l:max_l]
        # Load images
        images_tmp = []
        for f in tmp_path_list:
            tmp_im = Image.open(f)
            tmp_im.load() # Cumpulsory to avoid "Too many open files" issue
            images_tmp.append(tmp_im)
        # Process images
        processed_images_tmp = preprocessor(images_tmp)
        # Get new files list
        new_path_list_tmp = [os.path.relpath(f, old_directory) for f in tmp_path_list]
        new_path_list_tmp = [os.path.join(tmp_folder, f) for f in new_path_list_tmp]
        # Save in PNG format -> lossless compression ! DO NOT SAVE IN JPEG !
        new_path_list_tmp = [f"{os.path.splitext(f)[0]}.png" for f in new_path_list_tmp]
        # Save processed images
        for i in range(len(processed_images_tmp)):
            im = processed_images_tmp[i]
            im_path = new_path_list_tmp[i]
            # Check if folder exists !
            dst_dir_path = os.path.dirname(im_path)
            if not os.path.exists(dst_dir_path):
                os.makedirs(dst_dir_path)
            im.save(im_path, format='PNG')
        # Add new path to the main list
        new_path_list += new_path_list_tmp

    # Return
    return new_path_list


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--directory', default=None, required=True, help="Name of the test directory (actually a path relative to {{package_name}}-data)")
    # model_X should be the model's directory name: e.g. model_tfidf_svm_2019_12_05-12_57_18
    parser.add_argument('-m', '--model_dir', required=True, help="Name of the model to use (not a path, just the directory name)")
    parser.add_argument('--sep', default='{{default_sep}}', help="Separator to use with the .csv files.")
    parser.add_argument('--encoding', default="{{default_encoding}}", help="Encoding to use with the .csv files.")
    parser.add_argument('--force_cpu', dest='on_cpu', action='store_true', help="Whether to force training on CPU (and not GPU)")
    parser.set_defaults(on_cpu=False)
    args = parser.parse_args()
    # Check forced CPU usage
    if args.on_cpu:
        os.environ['CUDA_VISIBLE_DEVICES'] = ""
        logger.info("----------------------------")
        logger.info("CPU USAGE FORCED BY THE USER")
        logger.info("----------------------------")
    # Main
    main(directory=args.directory, model_dir=args.model_dir, sep=args.sep, encoding=args.encoding)



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import dill as pickle
import shutil
import numpy as np
import pandas as pd
from PIL import Image

from {{package_name}} import utils
from {{package_name}}.monitoring.model_explainer import LimeExplainer
from {{package_name}}.models_training.classifiers.model_cnn_classifier import ModelCnnClassifier


# Disable logging
import logging
logging.disable(logging.CRITICAL)


def remove_dir(path):
    if os.path.isdir(path): shutil.rmtree(path)


class ModelExplainerTest(unittest.TestCase):
    '''Main class to test model_explainer'''

    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)
        # Check if data folder exists
        data_path = utils.get_data_path()
        if not os.path.exists(data_path):
            os.mkdir(data_path)
        # Create a mock embedding
        fake_embedding = {'toto': [0.25, 0.90, 0.12], 'titi': [0.85, 0.12, 0.8], 'test': [0.5, 0.6, 0.1]}
        fake_path = os.path.join(data_path, 'fake_embedding.pkl')
        if os.path.exists(fake_path):
            os.remove(fake_path)
        with open(fake_path, 'wb') as f:
            pickle.dump(fake_embedding, f, pickle.HIGHEST_PROTOCOL)

    def test01_lime_explainer_nominal(self):
        '''Test of the Lime explainer'''

        # Model dir
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        # fake model_conf
        model_conf = {'preprocess_str': 'no_preprocess'}
        # Set vars
        data_path = os.path.join(os.getcwd(), 'test_data', '{{package_name}}-data', 'test_data_1')
        filenames = ['Birman_1.jpg', 'Birman_2.jpg', 'Birman_4.jpg', 'Bombay_1.png', 'Bombay_10.jpg',
                     'Bombay_3.jpg', 'shiba_inu_15.jpg', 'shiba_inu_30.jpg', 'shiba_inu_31.jpg',
                     'shiba_inu_34.jpg', 'subfolder/Birman_36.jpg', 'subfolder/Bombay_19.jpg']
        df_train_mono = pd.DataFrame({
            'file_class': ['cat', 'cat', 'cat', 'cat', 'cat', 'cat', 'dog', 'dog', 'dog', 'dog', 'cat', 'cat'],
            'file_path': [os.path.join(data_path, _) for _ in filenames],
        })
        df_train_multi = pd.DataFrame({
            'file_class': ['birman', 'birman', 'birman', 'bombay', 'bombay', 'bombay', 'shiba', 'shiba', 'shiba', 'shiba', 'birman', 'bombay'],
            'file_path': [os.path.join(data_path, _) for _ in filenames],
        })
        img = Image.open(df_train_mono['file_path'].values[0])

        # Mono Class
        model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2)
        model.fit(df_train_mono, df_valid=df_train_mono)
        explainer = LimeExplainer(model, model_conf)
        explanation = explainer.explain_instance(content=img, class_or_label_index=0)
        explanation = explainer.explain_instance(content=img, class_or_label_index=1)
        explanation = explainer.explain_instance(content=img, class_or_label_index=None)
        remove_dir(model_dir)

        # Multi classes
        model = ModelCnnClassifier(model_dir=model_dir, batch_size=2, epochs=2)
        model.fit(df_train_multi, df_valid=df_train_multi)
        explainer = LimeExplainer(model, model_conf)
        explanation = explainer.explain_instance(content=img, class_or_label_index=0)
        explanation = explainer.explain_instance(content=img, class_or_label_index=1)
        explanation = explainer.explain_instance(content=img, class_or_label_index=2)
        explanation = explainer.explain_instance(content=img, class_or_label_index=None)
        remove_dir(model_dir)

        # Check errors
        # Not a classifier
        class FakeModelClass1():
            def __init__(self):
                self.list_classes = ['a', 'b', 'c']
                self.model_type = 'regressor'
            def predict_proba(self):
                pass
        with self.assertRaises(ValueError):
            LimeExplainer(FakeModelClass1(), model_conf)
        # No predict_proba
        class FakeModelClass2():
            def __init__(self):
                self.list_classes = ['a', 'b', 'c']
                self.model_type = 'classifier'
        with self.assertRaises(TypeError):
            LimeExplainer(FakeModelClass2(), model_conf)
        # No list_classes
        class FakeModelClass3():
            def __init__(self):
                self.model_type = 'classifier'
            def predict_proba(self):
                pass
        with self.assertRaises(TypeError):
            LimeExplainer(FakeModelClass3(), model_conf)


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3

## Fonctions utiles pour le tutoriel Jules Verne
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
import math
import pandas as pd
from typing import List, Union


def text_to_sentences(text: str, nb_word_sentence: int = 10) -> List[str]:
    '''Transforms a text in sentences.

    Args:
        text (str): The text to cut in sentences
    Kwargs:
        nb_word_sentence (int): The number of words in a sentence
    Returns:
        list: A list of sentences
    '''
    text = re.sub(r'\s', ' ', text)
    # Change some punctuations to a whitespace
    for punctuation in [r'\!', r'\?', r'\;', r'\.', r'\,', r'\:']:
        text = re.sub(punctuation, ' ', text)
    # Get rid of superfluous whitespaces
    text = re.sub(' +', ' ', text)
    words_list = text.split(' ')
    list_sentences = []
    # Cut the text in sentences
    for i in range(math.ceil(len(words_list) / nb_word_sentence)):
        list_sentences.append(' '.join(words_list[i * nb_word_sentence: (i + 1) * nb_word_sentence]))
    return list_sentences


def df_texts_to_df_sentences(texts: Union[list, pd.Series], authors: Union[list, pd.Series], nb_word_sentence: int = 10) -> pd.DataFrame:
    '''Creates a dataframe containaing 'sentences' from a collection of texts with their corresponding authors.

    Args:
        texts (list, pd.Series): a collection of texts
        authors (list, pd.Series): the corresponding authors
    Kwargs:
        nb_word_sentence (int): The number of words in a sentence
    Returns:
        pd.DataFrame: A dataframe containing sentences extracted from the initial texts with their corresponding author
    '''
    list_sentences = []
    for text, author in zip(list(texts), list(authors)):
        # This function transforms a text in 'sentences'
        sentences = text_to_sentences(text, nb_word_sentence)
        list_sentences = list_sentences + [(sentence, author) for sentence in sentences]
    df_sentences = pd.DataFrame(list_sentences, columns=['sentence', 'author'])
    return df_sentences



#!/usr/bin/env python3

## Functions to remove white borders
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import tqdm
import logging
import numpy as np
from PIL import Image
from math import floor
from typing import Union


# Get logger
logger = logging.getLogger(__name__)


def remove_white_borders(images: list, image_ratio_strategy: Union[str, None] = None, image_ratio: float = 0.75, with_rotation: bool = True) -> list:
    '''Removes white border
    Also change the image ratio and rotate (if wanted) along largest dim. (i.e. portrait mode)

    Args:
        images (list): Images to be processed
    Kwargs:
        image_ratio_strategy (str): wanted strategy to apply new image_ratio
            - None: no change in image ratio
            - 'fill': add white borders on the smallest dimension
            - 'stretch': stretch the images such that they have the wanted ratio
        image_ratio (float): Wanted final image ratio (unused if image_ratio_strategy is None)
        with_rotation (bool): If the images must be rotated along largest dim. (i.e. portrait mode)
    Raises:
        ValueError: If image_ratio_strategy value is not a valid option ([None, 'fill', 'stretch'])
    Returns:
        list: Images sans les marges blanches
    '''
    if image_ratio_strategy not in [None, 'fill', 'stretch']:
        raise ValueError(f"image ratio strategy (image_ratio_strategy) '{image_ratio_strategy}' is not a valid option ([None, 'fill', 'stretch'])")
    # Get 'True' white
    # TODO : to be improved !
    true_white = _rgb2gray(np.array([255, 255, 255]))
    # Process each image, one by one
    results = []
    for i, im in enumerate(tqdm.tqdm(images)):
        # Remove white borders
        pixels = _rgb2gray(np.array(im))
        # x : horizontal
        # y : vertical
        first_x = _get_first_x(pixels, true_white)  # Left
        first_y = _get_first_y(pixels, true_white)  # Upper
        last_x = _get_last_x(pixels, true_white)  # Right
        last_y = _get_last_y(pixels, true_white)  # Lower
        # If first_x -1 -> no 'non-white' pixel, do nothing
        if first_x == -1:
            continue
        else:
            # Crop image (left, upper, right, lower)
            im = im.crop((first_x, first_y, last_x + 1, last_y + 1))
            # Rotate image if wanted
            if with_rotation:
                im = rotate_image_largest_dim(im)
            # Manage new ratio strategy
            if image_ratio_strategy == 'fill':  # fill with white borders to get correct format
                im = fill_with_white(im, image_ratio)
            elif image_ratio_strategy == 'stretch':
                im = stretch_image(im, image_ratio)
            # If None, do nothign
        # Update list
        results.append(im)
    # Return
    return results


def _rgb2gray(rgb: np.ndarray) -> np.ndarray:
    '''Gets gray value from an rgb pixel

    Args:
        rgb (np.ndarray): Image to be processed
    Returns:
        np.ndarray: New image
    '''
    return np.dot(rgb[..., :3], [0.2989, 0.5870, 0.1140])


def _get_first_y(pixels: np.ndarray, true_white: np.ndarray) -> int:
    '''Gets first non-white y (vertical) pixel

    Args:
        pixels (np.ndarray): Image to be processed
        true_white (np.ndarray): "True" white (to be improved)
    Returns:
        int: Pixel's index
    '''
    mins = np.amin(pixels, axis=1)  # Get min pix per line
    # Look for first row with non-white pixel(s)
    for i, pix in enumerate(mins):
        if not pix == true_white:
            return i
    # No match ? Return -1
    return -1


def _get_first_x(pixels: np.ndarray, true_white: np.ndarray) -> int:
    '''Gets first non-white x (horizontal) pixel

    Args:
        pixels (np.ndarray): Image to be processed
        true_white (np.ndarray): "True" white (to be improved)
    Returns:
        int: Pixel's index
    '''
    new_pixels = np.transpose(pixels)
    return _get_first_y(new_pixels, true_white)


def _get_last_y(pixels: np.ndarray, true_white: np.ndarray) -> int:
    '''Gets last non-white y (vertical) pixel

    Args:
        pixels (np.ndarray): Image to be processed
        true_white (np.ndarray): "True" white (to be improved)
    Returns:
        int: Pixel's index
    '''
    new_pixels = np.flip(pixels, 0)
    tmp_y = _get_first_y(new_pixels, true_white)
    if tmp_y == -1:
        return -1
    else:
        return pixels.shape[0] - 1 - tmp_y


def _get_last_x(pixels: np.ndarray, true_white: np.ndarray) -> int:
    '''Gets last non-white x (horizontal) pixel

    Args:
        pixels (np.ndarray): Image to be processed
        true_white (np.ndarray): "True" white (to be improved)
    Returns:
        int: Pixel's index
    '''
    new_pixels = np.transpose(np.flip(pixels, 1))
    tmp_x = _get_first_y(new_pixels, true_white)
    if tmp_x == -1:
        return -1
    else:
        return pixels.shape[1] - 1 - tmp_x


def rotate_image_largest_dim(im: Image) -> Image:
    '''Rotates an image along largest dim. (i.e. portrait mode)

    Args:
        im (Image): Image to be processed
    Returns:
        Image: Rotated image
    '''
    orientation = _get_orientation(im)
    if orientation != 0:
        im = im.rotate(orientation, expand=True)
    return im


def fill_with_white(im: Image, image_ratio: float) -> Image:
    '''Fills an image with white such that it respects a wanted ratio

    Args:
        im (Image): Image to be processed
    Kwargs:
        image_ratio (float): Wanted image ratio
    Returns:
        Image: Transformed image
    '''
    width, height = im.size
    ratio = width / height
    if ratio > image_ratio:
        # Increase height
        wanted_height = round(width / image_ratio)
        new_size = (width, wanted_height)
        old_size = (width, height)
        # Set new image
        new_im = Image.new("RGB", new_size, (255, 255, 255))
        # Fill it with the old image, centered
        # (use floor to ensure the old image fits into the new one)
        x_pos = 0
        y_pos = floor((new_size[1] - old_size[1]) / 2)
        new_im.paste(im, (x_pos, y_pos))
    elif ratio < image_ratio:
        # Increase width
        wanted_width = round(height * image_ratio)
        new_size = (wanted_width, height)
        old_size = (width, height)
        # Set new image
        new_im = Image.new("RGB", new_size, (255, 255, 255))
        # Fill it with the old image, centered
        # (use floor to ensure the old image fits into the new one)
        x_pos = floor((new_size[0] - old_size[0]) / 2)
        y_pos = 0
        new_im.paste(im, (x_pos, y_pos))
    else:  # Already correct ratio
        new_im = im
    return new_im


def stretch_image(im: Image, image_ratio: float) -> Image:
    '''Stretch an image such that it respects a wanted ratio

    Args:
        im (Image): Image to be processed
    Kwargs:
        image_ratio (float): Wanted image ratio
    Returns:
        Image: Transformed image
    '''
    width, height = im.size
    ratio = width / height
    if ratio > image_ratio:
        # Increase height
        wanted_height = round(width / image_ratio)
        new_size = (width, wanted_height)
        new_im = im.resize(new_size)
    elif ratio < image_ratio:
        # Increase width
        wanted_width = round(height * image_ratio)
        new_size = (wanted_width, height)
        new_im = im.resize(new_size)
    else:  # Already correct ratio
        new_im = im
    return new_im


def _get_orientation(im: Image) -> int:
    '''Gets an image orientation based on largest dimension

    Args:
        im (Image): Image to be processed
    Returns:
        int: Image orientation
    '''
    if im.size[0] > im.size[1]:
        return 90
    else:
        return 0


if __name__ == '__main__':
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import random
import tempfile
import numpy as np
import pandas as pd
from PIL import Image
from skimage import io

from {{package_name}}.models_training.object_detectors import utils_faster_rcnn

# Disable logging
import logging
logging.disable(logging.CRITICAL)

class UtilsFasterRCNNTests(unittest.TestCase):
    '''Main class to test all functions in utils_faster_rcnn.py'''
    
    @classmethod
    def setUpClass(cls):
        '''SetUp fonction'''
        cls.test_dir = os.path.dirname(os.path.abspath(__file__))

    def test01_RoiPoolingLayer(self):
        '''Test of the class utils_faster_rcnn.RoiPoolingLayer'''
        layer = utils_faster_rcnn.RoiPoolingLayer(2)

        features_maps = np.zeros(shape=(1, 10, 10, 1)) # Shape (batch_size, cols, rows, channels)
        rois = np.array([[[0, 0, 2, 2], [2, 2, 3, 3]]])

        result = layer([features_maps, rois])
        self.assertTrue(result.shape, (1, 2, 2, 2, 1))

    def test02_get_rpn_loss_regr(self):
        '''Test of the function utils_faster_rcnn.get_rpn_loss_regr'''
        loss = utils_faster_rcnn.get_rpn_loss_regr(1)

        y_true = np.ones(shape=(1, 10, 10, 2*4*1))
        y_pred = np.zeros(shape=(1, 10, 10, 4*1))

        result = loss(y_true, y_pred)
        self.assertAlmostEqual(result, 0.5, delta=0.1)

    def test02_get_rpn_loss_cls(self):
        '''Test of the function utils_faster_rcnn.get_rpn_loss_cls'''
        loss = utils_faster_rcnn.get_rpn_loss_cls(1)

        y_true = np.ones(shape=(1, 10, 10, 2*1))
        y_pred = np.zeros(shape=(1, 10, 10, 1))

        result = loss(y_true, y_pred)
        self.assertAlmostEqual(result.numpy(), 15.4, delta=0.1)

    def test03_get_class_loss_regr(self):
        '''Test of the function utils_faster_rcnn.get_class_loss_regr'''
        loss = utils_faster_rcnn.get_class_loss_regr(1)

        y_true = np.ones(shape=(1, 2, 2*4*1))
        y_pred = np.zeros(shape=(1, 2, 4*1))

        result = loss(y_true, y_pred)
        self.assertAlmostEqual(result, 0.5, delta=0.1)

    def test04_class_loss_cls(self):
        '''Test of the function utils_faster_rcnn.class_loss_cls'''
        y_true = [[0, 1, 0], [0, 0, 1]]
        y_pred = [[0.05, 0.95, 0], [0.1, 0.8, 0.1]]

        result = utils_faster_rcnn.class_loss_cls(y_true, y_pred)
        self.assertAlmostEqual(result.numpy(), 1.17, delta=0.1)

    def test05_get_custom_objects_faster_rcnn(self):
        '''Test of the function utils_faster_rcnn.get_custom_objects_faster_rcnn'''
        result = utils_faster_rcnn.get_custom_objects_faster_rcnn(1, 2)
        for key in ("RoiPoolingLayer", "rpn_loss_regr", "rpn_loss_cls", "class_loss_cls", "class_loss_regr"):
            self.assertIn(key, result)


#!/usr/bin/env python3

## Utils - Tools for training
# Copyright (C) <2018-2021>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Functions :
# - get_embedding -> Loads an embedding previously saved as a .pkl file
# - normal_split -> Splits a DataFrame into train and test sets
# - stratified_split -> Splits a DataFrame into train and test sets - Stratified strategy
# - hierarchical_split -> Splits a DataFrame into train and test sets - Hierarchical strategy
# - remove_small_classes -> Deletes under-represented classes
# - display_train_test_shape -> Displays the size of a train/test split
# - preprocess_model_multilabel -> Prepares a dataframe for a multi-labels model
# - load_pipeline -> Loads a pipeline from the pipelines folder
# - load_model -> Loads a model from a path
# - get_columns_pipeline -> Retrieves a pipeline wanted columns, and mandatory ones
# - apply_pipeline -> Applies a fitted pipeline to a dataframe
# - predict -> Gets predictions on a dataset
# - predict_with_proba -> Gets predictions with probabilities on a dataset
# - search_hp_cv -> Searches for hyperparameters


import os
import gc
import json
import math
import pprint
import shutil
import logging
import numpy as np
import pandas as pd
import dill as pickle
from datetime import datetime
from typing import Union, Tuple, Callable, Any, Dict, List

from sklearn.preprocessing import MultiLabelBinarizer
from sklearn.model_selection import train_test_split, KFold, StratifiedKFold

from .. import utils
from .model_class import ModelClass
from ..preprocessing import preprocess


# Get logger
logger = logging.getLogger(__name__)


def get_embedding(embedding_name: str = "cc.fr.300.pkl") -> Dict[str, np.ndarray]:
    '''Loads an embedding previously saved as a .pkl file

    Args:
        embedding_name (str): Name of the embedding file (actually a path relative to {{package_name}}-data)
    Raises:
        FileNotFoundError: If the embedding file does not exist in {{package_name}}-data
    Returns:
        dict: Loaded embedding
    '''
    logger.info("Reloading an embedding ...")

    # Manage path
    data_path = utils.get_data_path()
    embedding_path = os.path.join(data_path, embedding_name)
    if not os.path.exists(embedding_path):
        logger.error(f"The provided embedding file ({embedding_path}) does not exist.")
        logger.error("If you want to create one, you have to :")
        logger.error("    1. Download a fasttext embedding (e.g. French -> https://dl.fbaipublicfiles.com/fasttext/vectors-crawl/cc.fr.300.vec.gz")
        logger.error("    2. Use the 0_get_embedding_dict.py to generate the pkl. file")
        raise FileNotFoundError()

    # Load embedding indexes
    with open(embedding_path, 'rb') as f:
        embedding_indexes = pickle.load(f)

    # Return
    return embedding_indexes


def normal_split(df: pd.DataFrame, test_size: float = 0.25, seed: Union[int, None] = None) -> Tuple[pd.DataFrame, pd.DataFrame]:
    '''Splits a DataFrame into train and test sets

    Args:
        df (pd.DataFrame): Dataframe containing the data
    Kwargs:
        test_size (float): Proportion representing the size of the expected test set
        seed (int): random seed
    Raises:
        ValueError: If the object test_size is not between 0 and 1
    Returns:
        DataFrame: Train dataframe
        DataFrame: Test dataframe
    '''
    if not 0 <= test_size <= 1:
        raise ValueError('The object test_size must be between 0 and 1')

    # Normal split
    logger.info("Normal split")
    df_train, df_test = train_test_split(df, test_size=test_size, random_state=seed)

    # Display
    display_train_test_shape(df_train, df_test, df_shape=df.shape[0])

    # Return
    return df_train, df_test


def stratified_split(df: pd.DataFrame, col: Union[str, int], test_size: float = 0.25, seed: int = None) -> Tuple[pd.DataFrame, pd.DataFrame]:
    '''Splits a DataFrame into train and test sets - Stratified strategy

    Args:
        df (pd.DataFrame): Dataframe containing the data
        col (str or int): column on which to do the stratified split
    Kwargs:
        test_size (float): Proportion representing the size of the expected test set
        seed (int): Random seed
    Raises:
        ValueError: If the object test_size is not between 0 and 1
    Returns:
        DataFrame: Train dataframe
        DataFrame: Test dataframe
    '''
    if not 0 <= test_size <= 1:
        raise ValueError('The object test_size must be between 0 and 1')

    # Stratified split
    logger.info("Stratified split")
    df = remove_small_classes(df, col, min_rows=math.ceil(1 / test_size))  # minimum lines number per category to split
    df_train, df_test = train_test_split(df, stratify=df[col], test_size=test_size, random_state=seed)

    # Display
    display_train_test_shape(df_train, df_test, df_shape=df.shape[0])

    # Return
    return df_train, df_test


def hierarchical_split(df: pd.DataFrame, col: Union[str, int], test_size: float = 0.25, seed: int = None) -> Tuple[pd.DataFrame, pd.DataFrame]:
    '''Splits a DataFrame into train and test sets - Hierarchical strategy

    Args:
        df (pd.DataFrame): Dataframe containing the data
        col (str or int): column on which to do the hierarchical split
    Kwargs:
        test_size (float): Proportion representing the size of the expected test set
        seed (int): Random seed
    Raises:
        ValueError: If the object test_size is not between 0 and 1
    Returns:
        DataFrame: Train dataframe
        DataFrame: Test dataframe
    '''
    if not 0 <= test_size <= 1:
        raise ValueError('The object test_size must be between 0 and 1')

    # Hierarchical split
    logger.info("Hierarchical split")
    modalities_train, _ = train_test_split(df[col].unique(), test_size=test_size, random_state=seed)
    train_rows = df[col].isin(modalities_train)
    df_train = df[train_rows]
    df_test = df[~train_rows]

    # Display
    display_train_test_shape(df_train, df_test, df_shape=df.shape[0])

    # Return
    return df_train, df_test


def remove_small_classes(df: pd.DataFrame, col: Union[str, int], min_rows: int = 2) -> pd.DataFrame:
    '''Deletes the classes with small numbers of elements

    Args:
        df (pd.DataFrame): Dataframe containing the data
        col (str | int): Columns containing the classes
    Kwargs:
        min_rows (int): Minimal number of lines in the training set (default: 2)
    Raises:
        ValueError: If the object min_rows is not positive
    Returns:
        pd.DataFrame: New dataset
    '''
    if min_rows < 1:
        raise ValueError("The object min_rows must be positive")

    # Looking for classes with less than min_rows lines
    v_count = df[col].value_counts()
    classes_to_remove = list(v_count[v_count < min_rows].index.values)
    for cl in classes_to_remove:
        logger.warning(f"/!\\ /!\\ /!\\ Class {cl} has less than {min_rows} lines in the training set.")
        logger.warning("/!\\ /!\\ /!\\ This class is automatically removed from the dataset.")
    return df[~df[col].isin(classes_to_remove)]


def display_train_test_shape(df_train: pd.DataFrame, df_test: pd.DataFrame, df_shape: Union[int, None] = None) -> None:
    '''Displays the size of a train/test split

    Args:
        df_train (pd.DataFrame): Train dataset
        df_test (pd.DataFrame): Test dataset
    Kwargs:
        df_shape (int): Size of the initial dataset
    Raises:
        ValueError: If the object df_shape is not positive
    '''
    if df_shape is not None and df_shape < 1:
        raise ValueError("The object df_shape must be positive")

    # Process
    if df_shape is None:
        df_shape = df_train.shape[0] + df_test.shape[0]
    logger.info(f"There are {df_train.shape[0]} lines in the train dataset and {df_test.shape[0]} in the test dataset.")
    logger.info(f"{round(100 * df_train.shape[0] / df_shape, 2)}% of data are in the train set")
    logger.info(f"{round(100 * df_test.shape[0] / df_shape, 2)}% of data are in the test set")


def preprocess_model_multilabel(df: pd.DataFrame, y_col: Union[str, int], classes: Union[list, None] = None) -> Tuple[pd.DataFrame, list]:
    '''Prepares a dataframe for a multi-labels classification

    Args:
        df (pd.DataFrame): Training dataset
            This dataset must be preprocessed.
            Example:
                # Group by & apply tuple to y_col
                x_cols = [col for col in list(df.columns) if col != y_col]
                df = pd.DataFrame(df.groupby(x_cols)[y_col].apply(tuple))
        y_col (str or int): Name of the column to be used for training - y
    Kwargs:
        classes (list): List of classes to consider
    Returns:
        DataFrame: Dataframe for training
        list: List of 'y' columns
    '''
    # TODO: add possibility to have sparse output
    logger.info("Preprocess dataframe for multi-labels model")
    # Process
    logger.info("Preparing dataset for multi-labels format. Might take several minutes.")
    # /!\ The reset_index is compulsory in order to have the same indexes between df, and MLB transformed values
    df = df.reset_index(drop=True)
    # Apply MLB
    mlb = MultiLabelBinarizer(classes=classes)
    df = df.assign(**pd.DataFrame(mlb.fit_transform(df[y_col]), columns=mlb.classes_))
    # Return dataframe & y_cols (i.e. classes)
    return df, list(mlb.classes_)


def load_model(model_dir: str, is_path: bool = False, **kwargs) -> Tuple[Any, dict]:
    '''Loads a model from a path or a model name

    Args:
        model_dir (str): Name of the folder containing the model (e.g. model_autres_2019_11_07-13_43_19)
            It can also be an absolute path if is_path is set to True
    Kwargs:
        is_path (bool): If folder path instead of name (allows to load model from anywhere)
    Returns:
        ModelClass: The loaded model
        dict: The model configurations
    '''
    # Find model absolute path
    base_folder = None if is_path else utils.get_models_path()
    model_dir = utils.find_folder_path(model_dir, base_folder)

    # Load model
    model, model_conf = ModelClass.load_model(model_dir=model_dir, **kwargs)

    # Return model & its configs
    return model, model_conf


def predict(content: Union[str, list], model, model_conf: dict, **kwargs) -> list:
    '''Gets predictions of a model on a content

    Args:
        content (Union[str, list]): New content to be predicted
        model (ModelClass): Model to use
        model_conf (dict): Model configurations
    Returns:
        list: a list of strings (resp. tuples) in case of mono-label (resp. multi-labels) classification predictions
    '''
    if isinstance(content, str):
        content = [content]
        
    # Get preprocessor
    if 'preprocess_str' in model_conf.keys():
        preprocess_str = model_conf['preprocess_str']
    else:
        preprocess_str = 'no_preprocess'
    preprocessor = preprocess.get_preprocessor(preprocess_str)

    # Preprocess
    content = preprocessor(content)

    # Get prediction (some models need an iterable)
    predictions = model.predict(content)

    # Return predictions with inverse transform when relevant
    return model.inverse_transform(predictions)


def predict_with_proba(content: Union[str, list], model, model_conf: dict) -> Union[Tuple[List[str], List[float]], Tuple[List[tuple], List[tuple]]]:
    '''Gets predictions of a model on a content, with probabilities

    Args:
        content (Union[str, list]): New content to be predicted
        model (ModelClass): Model to use
        model_conf (dict): Model configurations
    Returns:
        MONO-LABEL CLASSIFICATION:
            List[str]: predictions
            List[float]: probabilities
        MULTI-LABELS CLASSIFICATION:
            List[tuple]: predictions
            List[tuple]: probabilities
    '''
    if isinstance(content, str):
        content = [content]

    # Get preprocessor
    if 'preprocess_str' in model_conf.keys():
        preprocess_str = model_conf['preprocess_str']
    else:
        preprocess_str = 'no_preprocess'
    preprocessor = preprocess.get_preprocessor(preprocess_str)

    # Preprocess
    content = preprocessor(content)

    # Get prediction (some models need an iterable)
    # predictions is a ndarray of shape (n_samples, n_classes)
    # probas is a ndarray of shape (n_samples, n_classes)
    predictions, probas = model.predict_with_proba(content)

    # Rework format :
    if model.multi_label:
        model_labels = np.array(model.list_classes)
        all_preds = [tuple(np.compress(content_pred, model_labels)) for content_pred in predictions]
        all_probs = [tuple(np.compress(content_pred, content_prob)) for content_pred, content_prob in zip(predictions, probas)]

    else:
        all_preds = model.inverse_transform(predictions)
        all_probs = probas.max(axis=1)

    # Return prediction & proba
    return all_preds, all_probs


def search_hp_cv(model_cls, model_params: dict, hp_params: dict, scoring_fn: Union[str, Callable],
                 kwargs_fit: dict, n_splits: int = 5):
    '''Searches for hyperparameters

    Args:
        model_cls (?): Class of models on which to do a hyperparameters search
        model_params (dict): Set of "fixed" parameters of the model (e.g. x_col, y_col).
            Must contain 'multi_label'.
        hp_params (dict): Set of "variable" parameters on which to do a hyperparameters search
        scoring_fn (str or func): Scoring function to maximize
            This function must take as input a dictionary containing metrics
            e.g. {'F1-Score': 0.85, 'Accuracy': 0.57, 'Precision': 0.64, 'Recall': 0.90}
        kwargs_fit (dict): Set of kwargs to input in the fit function
            Must contain 'x_train' and 'y_train'
    Kwargs:
        n_splits (int): Number of folds to use
    Raises:
        ValueError: If scoring_fn is not a known string
        ValueError: If multi_label is not a key in model_params
        ValueError: If x_train is not a key in kwargs_fit
        ValueError: If y_train is not a key in kwargs_fit
        ValueError: If model_params and hp_params share some keys
        ValueError: If hp_params values are not the same length
        ValueError: If the number of crossvalidation split is less or equal to 1
    Returns:
        ModelClass: best model to be "fitted" on the dataset
    '''
    list_known_scoring = ['accuracy', 'f1', 'precision', 'recall']

    #################
    # Manage errors
    #################

    if isinstance(scoring_fn, str) and scoring_fn not in list_known_scoring:
        raise ValueError(f"The input {scoring_fn} is not a known value for scoring_fn (known values : {list_known_scoring})")

    if 'multi_label' not in model_params.keys():
        raise ValueError("The key 'multi_label' must be present in the dictionary model_params")

    if 'x_train' not in kwargs_fit.keys():
        raise ValueError("The key 'x_train' must be present in the dictionary kwargs_fit")

    if 'y_train' not in kwargs_fit.keys():
        raise ValueError("The key 'y_train' must be present in the dictionary kwargs_fit")

    if any([k in hp_params.keys() for k in model_params.keys()]):
        # A key can't be "fixed" and "variable"
        raise ValueError("The dictionaries model_params and hp_params share at least one key")

    if len(set([len(_) for _ in hp_params.values()])) != 1:
        raise ValueError("The values of hp_params must have the same length")

    if n_splits <= 1:
        raise ValueError(f"The number of crossvalidation splits ({n_splits}) must be more than 1")

    #################
    # Manage scoring
    #################

    # Get scoring functions
    if scoring_fn == 'accuracy':
        scoring_fn = lambda x: x['Accuracy']
    elif scoring_fn == 'f1':
        scoring_fn = lambda x: x['F1-Score']
    elif scoring_fn == 'precision':
        scoring_fn = lambda x: x['Precision']
    elif scoring_fn == 'recall':
        scoring_fn = lambda x: x['Recall']

    #################
    # Manage x_train & y_train format
    #################

    if not isinstance(kwargs_fit['x_train'], (pd.DataFrame, pd.Series)):
        kwargs_fit['x_train'] = pd.Series(kwargs_fit['x_train'].copy())

    if not isinstance(kwargs_fit['y_train'], (pd.DataFrame, pd.Series)):
        kwargs_fit['y_train'] = pd.Series(kwargs_fit['y_train'].copy())

    #################
    # Process
    #################

    # Loop on hyperparameters
    nb_search = len(list(hp_params.values())[0])
    logger.info("Beginning of hyperparameters search")
    logger.info(f"Number of model fits : {nb_search} (search number) x {n_splits} (CV splits number) = {nb_search * n_splits}")

    # DataFrame for stocking metrics :
    metrics_df = pd.DataFrame(columns=['index_params', 'index_fold', 'Score', 'Accuracy', 'F1-Score', 'Precision', 'Recall'])
    for i in range(nb_search):

        # Display informations
        logger.info(f"Search n°{i + 1}")
        tmp_hp_params = {k: v[i] for k, v in hp_params.items()}
        logger.info("Tested hyperparameters : ")
        logger.info(pprint.pformat(tmp_hp_params))

        # Get folds (shuffle recommended since the classes could be ordered)
        if model_params['multi_label']:
            k_fold = KFold(n_splits=n_splits, shuffle=True)  # Can't stratify on multi-labels
        else:
            k_fold = StratifiedKFold(n_splits=n_splits, shuffle=True)

        # Process each fold
        for j, (train_index, valid_index) in enumerate(k_fold.split(kwargs_fit['x_train'], kwargs_fit['y_train'])):
            logger.info(f"Search n°{i + 1}/{nb_search} - fit n°{j + 1}/{n_splits}")
            # get tmp x, y
            x_train, x_valid = kwargs_fit['x_train'].iloc[train_index], kwargs_fit['x_train'].iloc[valid_index]
            y_train, y_valid = kwargs_fit['y_train'].iloc[train_index], kwargs_fit['y_train'].iloc[valid_index]
            # Get tmp model
            # Manage model_dir
            tmp_model_dir = os.path.join(utils.get_models_path(), datetime.now().strftime("tmp_%Y_%m_%d-%H_%M_%S"))
            # The next line prioritize the last dictionary
            # We force a temporary folder and a low save level (we only want the metrics)
            model_tmp = model_cls(**{**model_params, **tmp_hp_params, **{'model_dir': tmp_model_dir, 'level_save': 'LOW'}})
            # Setting log level to ERROR
            model_tmp.logger.setLevel(logging.ERROR)
            # Let's fit ! (priority to the last dictionary)
            model_tmp.fit(**{**kwargs_fit, **{'x_train': x_train, 'y_train': y_train, 'x_valid': x_valid, 'y_valid': y_valid}})
            # Let's predict !
            y_pred = model_tmp.predict(x_valid)
            # Get metrics !
            metrics_func = model_tmp.get_metrics_simple_multilabel if model_tmp.multi_label else model_tmp.get_metrics_simple_monolabel
            metrics_tmp = metrics_func(y_valid, y_pred)
            metrics_tmp = metrics_tmp[metrics_tmp.Label == "All"].copy()  # Add .copy() to avoid pandas settingwithcopy
            metrics_tmp["Score"] = scoring_fn(metrics_tmp.iloc[0].to_dict())  # type: ignore
            metrics_tmp["index_params"] = i
            metrics_tmp["index_fold"] = j
            metrics_tmp = metrics_tmp[metrics_df.columns]  # Keeping only the necessary columns
            metrics_df = pd.concat([metrics_df, metrics_tmp], ignore_index=True)
            # Delete the temporary model : the final model must be refitted on the whole dataset
            del model_tmp
            gc.collect()
            shutil.rmtree(tmp_model_dir)
        # Display score
        logger.info(f"Score for search n°{i + 1}: {metrics_df[metrics_df['index_params'] == i]['Score'].mean()}")

    # Metric agregation for all the folds
    metrics_df = metrics_df.join(metrics_df[['index_params', 'Score']].groupby('index_params').mean().rename({'Score': 'mean_score'}, axis=1), on='index_params', how='left')

    # Select the set of parameters with the best mean score (on the folds)
    best_index = metrics_df[metrics_df.mean_score == metrics_df.mean_score.max()]["index_params"].values[0]
    best_params = {k: v[best_index] for k, v in hp_params.items()}
    logger.info(f"Best results for the set of parameter n°{best_index + 1}: {pprint.pformat(best_params)}")

    # Instanciation of a model with the best parameters
    best_model = model_cls(**{**model_params, **best_params})

    # Save the metrics report of the hyperparameters search and the tested parameters
    csv_path = os.path.join(best_model.model_dir, "hyper_params_results.csv")
    metrics_df.to_csv(csv_path, sep='{{default_sep}}', index=False, encoding='{{default_encoding}}')
    json_data = {
        'model_params': model_params,
        'scoring_fn': pickle.source.getsourcelines(scoring_fn)[0],
        'n_splits': n_splits,
        'hp_params_set': {i: {k: v[i] for k, v in hp_params.items()} for i in range(nb_search)},
    }
    json_path = os.path.join(best_model.model_dir, "hyper_params_tested.json")
    with open(json_path, 'w', encoding='{{default_encoding}}') as f:
        json.dump(json_data, f, indent=4, cls=utils.NpEncoder)

    # TODO: We are forced to reset the logging level which is linked to the class
    best_model.logger.setLevel(logging.getLogger('{{package_name}}').getEffectiveLevel())

    # Return model to be fitted
    return best_model


if __name__ == '__main__':
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest

# Utils libs
import os
import json
import shutil
import numpy as np
import pandas as pd

from {{package_name}} import utils
from {{package_name}}.models_training.regressors.model_regressor import ModelRegressorMixin 
from {{package_name}}.models_training.regressors.models_sklearn.model_gbt_regressor import ModelGBTRegressor
from {{package_name}}.models_training.regressors.models_sklearn.model_sgd_regressor import ModelSGDRegressor
from {{package_name}}.models_training.classifiers.models_sklearn.model_sgd_classifier import ModelSGDClassifier
from {{package_name}}.models_training.regressors.model_aggregation_regressor import ModelAggregationRegressor, median_predict, mean_predict

# Disable logging
import logging
logging.disable(logging.CRITICAL)


def remove_dir(path):
    if os.path.isdir(path): shutil.rmtree(path)

models_path = utils.get_models_path()

def remove_dir_model(model, model_dir):
    for sub_model in model.sub_models:
        remove_dir(os.path.join(models_path, os.path.split(sub_model['model'].model_dir)[-1]))
    remove_dir(model_dir)

class MockModel(ModelRegressorMixin):

    def __init__(self, dict_predictions, model_name):
        self.dict_predictions = dict_predictions.copy()
        self.trained = True
        self.nb_fit = 1
        self.model_name = model_name
        self.model_dir = os.path.join('false_path',f'{model_name}')
        self.model_type = 'regressor'

    def predict(self, x_test, **kwargs):
        return np.array([self.dict_predictions[tuple(x)] for x in x_test])

dict_predictions_1 = {(1, 2): 1, (1, 3): 2, (1, 4): 3, (2,1): 2, (2,2): 5}
dict_predictions_2 = {(1, 2): -3, (1, 3): 3, (1, 4): 4, (2,1): 4, (2,2): 3}
dict_predictions_3 = {(1, 2): 4.1, (1, 3): 4.4, (1, 4): 5.2, (2,1): 3.1, (2,2): -23}
dict_predictions_4 = {(1, 2): 2, (1, 3): 1, (1, 4): -2, (2,1): 2, (2,2): 1}
dict_predictions_5 = {(1, 2): 6, (1, 3): -1, (1, 4): 0, (2,1): 1, (2,2): 2}
list_dict_predictions = [dict_predictions_1, dict_predictions_2, dict_predictions_3, dict_predictions_4, dict_predictions_5]

x_test = np.array([[1, 2], [1, 3], [1, 4], [2, 1], [2, 2]])

mock_target_mean = np.array([np.mean([dict_prediction[tuple(x)] for dict_prediction in list_dict_predictions]) for x in x_test])
mock_target_median = np.array([np.median([dict_prediction[tuple(x)] for dict_prediction in list_dict_predictions]) for x in x_test])

mock_model_1 = MockModel(dict_predictions_1, 'model_1')
mock_model_2 = MockModel(dict_predictions_2, 'model_2')
mock_model_3 = MockModel(dict_predictions_3, 'model_3')
mock_model_4 = MockModel(dict_predictions_4, 'model_4')
mock_model_5 = MockModel(dict_predictions_5, 'model_5')
list_mock_model = [mock_model_1, mock_model_2, mock_model_3, mock_model_4, mock_model_5]

target_get_predictions = np.array([[model.predict(x_test)[i] for model in list_mock_model] for i, x in enumerate(x_test)])


class ModelAggregationRegressorTests(unittest.TestCase):
    '''Main class to test model_aggregation_regressor'''

    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

    # Create and save a ModelGBTRegressor and a ModelSGDRegressor models
    def create_models(self, gbt_param: dict = None, sgd_param: dict = None):
        model_path = utils.get_models_path()
        model_dir_gbt = os.path.join(model_path, 'model_test_123456789_gbt')
        model_dir_sgd = os.path.join(model_path, 'model_test_123456789_sgd')
        remove_dir(model_dir_gbt)
        remove_dir(model_dir_sgd)

        gbt_param = {} if gbt_param is None else gbt_param
        sgd_param = {} if sgd_param is None else sgd_param

        gbt = ModelGBTRegressor(model_dir=model_dir_gbt, **gbt_param)
        sgd = ModelSGDRegressor(model_dir=model_dir_sgd, **sgd_param)

        gbt.save()
        sgd.save()
        gbt_name = os.path.split(gbt.model_dir)[-1]
        sgd_name = os.path.split(sgd.model_dir)[-1]
        return gbt, sgd, gbt_name, sgd_name

    def test01_model_aggregation_regressor_init(self):
        '''Test of {{package_name}}.models_training.model_aggregation_regressor.ModelAggregationRegressor.__init__'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        model_name = 'test_model_name'

        x_train = pd.DataFrame.from_dict({'0':{'1': 1.2, '2':3.2}, '1':{'1': 1.1, '2':-3.2}}, orient='index')
        y_train = pd.Series([0.1, 0.2])

        def test_init_partial(model, model_name, model_dir, list_model_names, trained):
            self.assertTrue(os.path.isdir(model_dir))
            self.assertEqual(model.model_name, model_name)
            self.assertTrue(isinstance(model.sub_models, list))
            self.assertEqual([sub_model['name'] for sub_model in model.sub_models], list_model_names)
            self.assertTrue(isinstance(model.sub_models[0]['model'], ModelGBTRegressor))
            self.assertTrue(isinstance(model.sub_models[1]['model'], ModelSGDRegressor))
            self.assertEqual(model.trained, trained)

            # We test display_if_gpu_activated and _is_gpu_activated just by calling them
            model.display_if_gpu_activated()
            self.assertTrue(isinstance(model._is_gpu_activated(), bool))

        ############################################
        # Init., test all parameters
        ############################################

        # list_models = [model, model]
        # aggregation_function: median_predict
        # using_proba
        # not multi_label
        gbt, sgd, gbt_name, sgd_name = self.create_models()
        gbt.fit(x_train, y_train)
        list_models = [gbt, sgd]
        list_model_names = [gbt_name, sgd_name]
        model = ModelAggregationRegressor(model_dir=model_dir, model_name=model_name, list_models=list_models, aggregation_function='median_predict')
        test_init_partial(model, model_name, model_dir, list_model_names, trained=False)
        self.assertEqual(median_predict.__code__.co_code, model.aggregation_function.__code__.co_code)
        remove_dir_model(model, model_dir)

        # list_models = [model_name, model_name]
        # aggregation_function: mean_predict
        # not using_proba
        # not multi_label
        gbt, sgd, gbt_name, sgd_name = self.create_models()
        list_models = [gbt_name, sgd_name]
        list_model_names = [gbt_name, sgd_name]
        model = ModelAggregationRegressor(model_dir=model_dir, model_name=model_name, list_models=list_models, aggregation_function='mean_predict')
        test_init_partial(model, model_name, model_dir, list_model_names, trained=False)
        self.assertEqual(mean_predict.__code__.co_code, model.aggregation_function.__code__.co_code)
        remove_dir_model(model, model_dir)

        # list_models = [model_name, model]
        # aggregation_function: median_predict
        # not using_proba
        # multi_label
        gbt, sgd, gbt_name, sgd_name = self.create_models()
        list_models = [gbt_name, sgd]
        list_model_names = [gbt_name, sgd_name]
        model = ModelAggregationRegressor(model_dir=model_dir, model_name=model_name, list_models=list_models, aggregation_function='median_predict')
        test_init_partial(model, model_name, model_dir, list_model_names, trained=False)
        self.assertEqual(median_predict.__code__.co_code, model.aggregation_function.__code__.co_code)
        remove_dir_model(model, model_dir)

        # list_models = [model_name, model]
        # aggregation_function: Callable
        # not using_proba
        # not multi_label

        # This function is a copy of median_predict function
        def function_test(predictions):
            return np.median(predictions)

        gbt, sgd, gbt_name, sgd_name = self.create_models()
        gbt.fit(x_train, y_train)
        gbt.save()
        sgd.fit(x_train, y_train)
        list_models = [gbt_name, sgd]
        list_model_names = [gbt_name, sgd_name]
        model = ModelAggregationRegressor(model_dir=model_dir, model_name=model_name, list_models=list_models, aggregation_function=function_test)
        test_init_partial(model, model_name, model_dir, list_model_names, trained=True)
        self.assertEqual(function_test.__code__.co_code, model.aggregation_function.__code__.co_code)
        remove_dir_model(model, model_dir)

        ############################################
        # Error
        ############################################

        # if the object aggregation_function is a str but not found in the dictionary dict_aggregation_function
        with self.assertRaises(ValueError):
            model = ModelAggregationRegressor(model_dir=model_dir, aggregation_function='1234')
        remove_dir(model_dir)

        # The classifier and regressor models cannot be combined in list models
        gbt, sgd, _, _ = self.create_models()
        model_dir_classifier = os.path.join(utils.get_models_path(), 'model_test_123456789_classifier')
        sgd_classifier = ModelSGDClassifier(model_dir=model_dir_classifier)
        list_models = [gbt, sgd_classifier]
        with self.assertRaises(ValueError):
            model = ModelAggregationRegressor(model_dir=model_dir, list_models=list_models)
        remove_dir_model(model, model_dir)
        remove_dir(sgd.model_dir)
        remove_dir(sgd_classifier.model_dir)

    def test02_model_aggregation_setter_aggregation_function(self):
        '''Test of {{package_name}}.models_training.model_aggregation_regressor.ModelAggregationRegressor.aggregation_function'''
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        model = ModelAggregationRegressor(model_dir=model_dir, multi_label=False)
        model.aggregation_function = 'mean_predict'
        self.assertEqual(model.aggregation_function.__code__.co_code, mean_predict.__code__.co_code)
        model.aggregation_function = 'median_predict'
        self.assertEqual(model.aggregation_function.__code__.co_code, median_predict.__code__.co_code)
        # str not in the dictionary
        with self.assertRaises(ValueError):
            model.aggregation_function = 'coucou'
        remove_dir(model_dir)

        # local aggregation function
        # This function is a copy of median_predict function
        def function_test(predictions: np.ndarray) -> np.float64:
            '''Returns the mean of the predictions of each model

            Args:
                predictions (np.ndarray) : The array containing the predictions of each models (shape (n_models))
            Return:
                (np.float64) : The mean of the predictions
            '''
            return np.mean(predictions)

        model = ModelAggregationRegressor(model_dir=model_dir)
        model.aggregation_function = function_test
        self.assertEqual(model.aggregation_function.__code__.co_code, function_test.__code__.co_code)
        remove_dir(model_dir)

    def test03_model_aggregation_setter_sub_models(self):
        '''Test of {{package_name}}.models_training.model_aggregation_regressor.ModelAggregationRegressor.sub_models'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        def check_sub_models(sub_models, list_models_name):
            self.assertTrue(isinstance(sub_models[0]['model'], ModelGBTRegressor))
            self.assertTrue(isinstance(sub_models[1]['model'], ModelSGDRegressor))
            self.assertEqual(len(sub_models), len(list_models))
            self.assertEqual([sub_model['name'] for sub_model in sub_models], list_models_name)

        # list_models = [model, model]
        gbt, sgd, gbt_name, sgd_name = self.create_models()
        list_models = [gbt, sgd]
        list_models_name = [gbt_name, sgd_name]
        model = ModelAggregationRegressor(model_dir=model_dir)
        model.sub_models = list_models
        check_sub_models(model.sub_models, list_models_name)
        remove_dir_model(model, model_dir)

        # list_models = [model_name, model_name]
        gbt, sgd, gbt_name, sgd_name = self.create_models()
        list_models = [gbt_name, sgd_name]
        list_models_name = [gbt_name, sgd_name]
        model = ModelAggregationRegressor(model_dir=model_dir)
        model.sub_models = list_models
        check_sub_models(model.sub_models, list_models_name)
        remove_dir_model(model, model_dir)

        # list_models = [model_name, model]
        gbt, sgd, gbt_name, sgd_name = self.create_models()
        list_models = [gbt_name, sgd]
        list_models_name = [gbt_name, sgd_name]
        model = ModelAggregationRegressor(model_dir=model_dir)
        model.sub_models = list_models
        check_sub_models(model.sub_models, list_models_name)
        remove_dir_model(model, model_dir)

    def test04_model_aggregation_regressor_check_trained(self):
        '''Test of {{package_name}}.models_training.model_aggregation_regressor.ModelAggregationRegressor._check_trained'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # set vars
        x_train = pd.DataFrame({'col_1': [-5, -1, 0, -2, 2, -6, 3] * 10, 'col_2': [2, -1, -8, 2, 3, 12, 2] * 10})
        y_train_int = pd.Series([0, 0, 0, 2, 1, 1, 1] * 10)

        # int
        gbt, sgd, _, _ = self.create_models()
        model = ModelAggregationRegressor(model_dir=model_dir)
        self.assertFalse(model._check_trained())
        gbt.fit(x_train, y_train_int)
        sgd.fit(x_train, y_train_int)
        model.sub_models = [gbt, sgd]
        self.assertTrue(model._check_trained())
        remove_dir_model(model, model_dir)

        # not trained
        gbt, sgd, _, _ = self.create_models()
        model = ModelAggregationRegressor(model_dir=model_dir)
        self.assertFalse(model._check_trained())
        model.sub_models = [gbt, sgd]
        self.assertFalse(model._check_trained())
        remove_dir_model(model, model_dir)

        gbt, sgd, _, _ = self.create_models()
        model = ModelAggregationRegressor(model_dir=model_dir)
        self.assertFalse(model._check_trained())
        sgd.fit(x_train, y_train_int)
        model.sub_models = [gbt, sgd]
        self.assertFalse(model._check_trained())
        remove_dir_model(model, model_dir)
        remove_dir(gbt.model_dir)
        remove_dir(sgd.model_dir)

    def test05_model_aggregation_regressor_fit(self):
        '''Test of {{package_name}}.models_training.model_aggregation_regressor.ModelAggregationRegressor.fit'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        def check_not_trained(model):
            self.assertFalse(model.trained)
            self.assertEqual(model.nb_fit, 0)

        def check_trained(model):
            self.assertTrue(model.trained)
            self.assertEqual(model.nb_fit, 1)
            for sub_model in model.sub_models:
                self.assertTrue(sub_model['model'].trained)
                self.assertEqual(sub_model['model'].nb_fit, 1)

        # Set vars
        x_train = pd.DataFrame({'col_1': [-5, -1, 0, -2, 2, -6, 3] * 10, 'col_2': [2, -1, -8, 2, 3, 12, 2] * 10})
        y_train = pd.Series([1, 1, 1, 2, 3, 3, 3] * 10)

        # Not trained
        gbt, sgd, _, _ = self.create_models()
        model = ModelAggregationRegressor(model_dir=model_dir, list_models=[gbt, sgd])
        check_not_trained(model)
        model.fit(x_train, y_train)
        check_trained(model)
        remove_dir_model(model, model_dir)

        # Some model trained
        gbt, sgd, _, _ = self.create_models()
        gbt.fit(x_train, y_train)
        model = ModelAggregationRegressor(model_dir=model_dir, list_models=[gbt, sgd])
        check_not_trained(model)
        model.fit(x_train, y_train)
        check_trained(model)
        remove_dir_model(model, model_dir)

    def test06_model_aggregation_regressor_predict(self):
        '''Test of {{package_name}}.models_training.model_aggregation_regressor.ModelAggregationRegressor.predict'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        def test_method(model, x_test, target_predict):
            preds = model.predict(x_test)
            self.assertEqual(preds.shape, target_predict.shape)
            for i in range(len(preds)):
                self.assertAlmostEqual(preds[i], target_predict[i], places=4)

        # median_predict
        model = ModelAggregationRegressor(model_dir=model_dir, list_models=list_mock_model, aggregation_function='median_predict')
        test_method(model, x_test, target_predict=mock_target_median)
        remove_dir(model_dir)

        # mean_predict
        model = ModelAggregationRegressor(model_dir=model_dir, list_models=list_mock_model, aggregation_function='mean_predict')
        test_method(model, x_test, target_predict=mock_target_mean)
        remove_dir(model_dir)

        # This function is a copy of median_predict function
        def function_test(predictions):
            return np.median(predictions)

        # callable
        model = ModelAggregationRegressor(model_dir=model_dir, list_models=list_mock_model, aggregation_function=function_test)
        test_method(model, x_test, target_predict=mock_target_median)
        remove_dir(model_dir)

        ############################################
        # Error
        ############################################

        x_train = pd.DataFrame({'col_1': [-5, -1, 0, -2, 2, -6, 3] * 10, 'col_2': [2, -1, -8, 2, 3, 12, 2] * 10})
        y_train = pd.Series([1, 1, 1, 2, 3, 3, 3] * 10)

        # Model needs to be fitted
        gbt, sgd, _, _ = self.create_models()
        model = ModelAggregationRegressor(model_dir=model_dir, list_models=[gbt, sgd])
        with self.assertRaises(AttributeError):
            model.predict(x_train)
        remove_dir_model(model, model_dir)

        # No return_proba
        gbt, sgd, _, _ = self.create_models()
        model = ModelAggregationRegressor(model_dir=model_dir, list_models=[gbt, sgd])
        model.fit(x_train, y_train)
        with self.assertRaises(ValueError):
            model.predict(x_train, return_proba=True)
        remove_dir_model(model, model_dir)

    def test07_model_aggregation_regressor_predict_proba(self):
        '''Test of {{package_name}}.models_training.model_aggregation_regressor.ModelAggregationRegressor.predict_proba'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        x_train = pd.DataFrame({'col_1': [-5, -1, 0, -2, 2, -6, 3] * 10, 'col_2': [2, -1, -8, 2, 3, 12, 2] * 10})
        y_train = pd.Series([1, 1, 1, 2, 3, 3, 3] * 10)

        gbt, sgd, _, _ = self.create_models()
        model = ModelAggregationRegressor(model_dir=model_dir, list_models=[gbt, sgd])
        model.fit(x_train, y_train)
        with self.assertRaises(ValueError):
            model.predict_proba(x_train)
        remove_dir_model(model, model_dir)

    def test08_model_aggregation_regressor_get_predictions(self):
        '''Test of {{package_name}}.models_training.model_aggregation_regressor.ModelAggregationRegressor._predict_sub_models'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        model = ModelAggregationRegressor(model_dir=model_dir, list_models=list_mock_model)
        preds = model._predict_sub_models(x_test)
        self.assertTrue(isinstance(preds, np.ndarray))
        self.assertEqual(target_get_predictions.shape, preds.shape)
        for i in range(target_get_predictions.shape[0]):
            for j in range(target_get_predictions.shape[1]):
                self.assertAlmostEqual(target_get_predictions[i][j], preds[i][j])
        remove_dir(model_dir)

        # Model needs to be fitted
        model = ModelAggregationRegressor(model_dir=model_dir)
        with self.assertRaises(AttributeError):
            model._predict_sub_models('test')
        remove_dir(model_dir)

    def test09_median_predict(self):
        '''Test of {{package_name}}.models_training.model_aggregation_regressor.median_predict'''

        # 3 models int
        preds = np.array([3, 4, 4])
        self.assertAlmostEqual(median_predict(preds), np.median(preds))
        # 3 models float
        preds = np.array([3, 3.1, 4])
        self.assertAlmostEqual(median_predict(preds), np.median(preds))
        # 1 model int
        preds = np.array([5])
        self.assertAlmostEqual(median_predict(preds), np.median(preds))
        # 1 model float
        preds = np.array([5.2])
        self.assertAlmostEqual(median_predict(preds), np.median(preds))


    def test10_mean_predict(self):
        '''Test of {{package_name}}.models_training.model_aggregation_regressor.mean_predict'''

        # 3 models int
        preds = np.array([3, 4, 4])
        self.assertAlmostEqual(mean_predict(preds), np.mean(preds))
        # 3 models float
        preds = np.array([3, 3.1, 4])
        self.assertAlmostEqual(mean_predict(preds), np.mean(preds))
        # 1 model int
        preds = np.array([5])
        self.assertAlmostEqual(mean_predict(preds), np.mean(preds))
        # 1 model float
        preds = np.array([5.2])
        self.assertAlmostEqual(mean_predict(preds), np.mean(preds))

    def test11_model_aggregation_regressor_save(self):
        '''Test of {{package_name}}.models_training.model_aggregation_regressor.ModelAggregationRegressor.save'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        x_train = pd.DataFrame({'col_1': [-5, -1, 0, -2, 2, -6, 3] * 10, 'col_2': [2, -1, -8, 2, 3, 12, 2] * 10})
        y_train = pd.Series([0.1, 0.5, 0.2, -2.9, 1.1, 3.2, 1.0] * 10)

        gbt, sgd, gbt_name, sgd_name = self.create_models()
        model = ModelAggregationRegressor(model_dir=model_dir, list_models=[gbt, sgd])
        model.fit(x_train, y_train)
        model.save(json_data={'test': 10})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"aggregation_function.pkl")))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"{model.model_name}.pkl")))

        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='utf-8') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 10)
        set_attributes_in_config = {'package_version', 'model_name', 'model_dir', 'trained', 'nb_fit', 
                                    'x_col', 'y_col', 'level_save', 'librairie'}
        set_attributes_in_config_tot = set_attributes_in_config.union({'list_models_name'})
        self.assertTrue(set_attributes_in_config_tot.issubset(set(configs.keys())))
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertEqual(configs['librairie'], None)

        for sub_model in model.sub_models:
            self.assertTrue(os.path.exists(os.path.join(sub_model['model'].model_dir, 'configurations.json')))
            self.assertTrue(os.path.exists(os.path.join(sub_model['model'].model_dir, f"{sub_model['model'].model_name}.pkl")))
            with open(os.path.join(sub_model['model'].model_dir, 'configurations.json'), 'r', encoding='utf-8') as f:
                configs = json.load(f)
            self.assertTrue(set_attributes_in_config.issubset(set(configs.keys())))
            self.assertEqual(configs['package_version'], utils.get_package_version())

        self.assertTrue(os.path.exists(os.path.join(model_dir, "model_upload_instructions.md")))
        with open(os.path.join(model_dir, "model_upload_instructions.md"), 'r') as read_obj:
            text = read_obj.read()
            self.assertEqual(text[0:20], "/!\\/!\/!\\/!\/!\\   The aggregation model is a special model, please ensure that"[0:20])
        remove_dir_model(model, model_dir)

        # Save each trained and unsaved model
        gbt, sgd, gbt_name, sgd_name = self.create_models()
        gbt.fit(x_train, y_train)
        model = ModelAggregationRegressor(model_dir=model_dir, list_models=[gbt, sgd])
        model.fit(x_train, y_train)
        model.save(json_data={'test': 10})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"aggregation_function.pkl")))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"{model.model_name}.pkl")))

        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='utf-8') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 10)
        set_attributes_in_config = {'package_version', 'model_name', 'model_dir', 'trained', 'nb_fit', 
                                    'x_col', 'y_col', 'level_save', 'librairie'}
        set_attributes_in_config_tot = set_attributes_in_config.union({'list_models_name'})
        self.assertTrue(set_attributes_in_config_tot.issubset(set(configs.keys())))
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertEqual(configs['librairie'], None)

        for sub_model in model.sub_models:
            self.assertTrue(os.path.exists(os.path.join(sub_model['model'].model_dir, 'configurations.json')))
            self.assertTrue(os.path.exists(os.path.join(sub_model['model'].model_dir, f"{sub_model['model'].model_name}.pkl")))
            with open(os.path.join(sub_model['model'].model_dir, 'configurations.json'), 'r', encoding='utf-8') as f:
                configs = json.load(f)
            self.assertTrue(set_attributes_in_config.issubset(set(configs.keys())))
            self.assertEqual(configs['package_version'], utils.get_package_version())

        self.assertTrue(os.path.exists(os.path.join(model_dir, "model_upload_instructions.md")))
        with open(os.path.join(model_dir, "model_upload_instructions.md"), 'r') as read_obj:
            text = read_obj.read()
            self.assertEqual(text[0:20], "/!\\/!\/!\\/!\/!\\   The aggregation model is a special model, please ensure that"[0:20])
        remove_dir_model(model, model_dir)

        # Same thing with a local function
        # This function is a copy of median_predict function
        def function_test(predictions):
            return np.median(predictions)

        gbt, sgd, gbt_name, sgd_name = self.create_models()
        model = ModelAggregationRegressor(model_dir=model_dir, list_models=[gbt, sgd], aggregation_function=function_test)
        model.fit(x_train, y_train)
        model.save(json_data={'test': 10})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"aggregation_function.pkl")))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"{model.model_name}.pkl")))

        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='utf-8') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 10)
        set_attributes_in_config = {'package_version', 'model_name', 'model_dir', 'trained', 'nb_fit', 
                                    'x_col', 'y_col', 'level_save', 'librairie'}
        set_attributes_in_config_tot = set_attributes_in_config.union({'list_models_name'})
        self.assertTrue(set_attributes_in_config_tot.issubset(set(configs.keys())))
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertEqual(configs['librairie'], None)

        for sub_model in model.sub_models:
            self.assertTrue(os.path.exists(os.path.join(sub_model['model'].model_dir, 'configurations.json')))
            self.assertTrue(os.path.exists(os.path.join(sub_model['model'].model_dir, f"{sub_model['model'].model_name}.pkl")))
            with open(os.path.join(sub_model['model'].model_dir, 'configurations.json'), 'r', encoding='utf-8') as f:
                configs = json.load(f)
            self.assertTrue(set_attributes_in_config.issubset(set(configs.keys())))
            self.assertEqual(configs['package_version'], utils.get_package_version())

        self.assertTrue(os.path.exists(os.path.join(model_dir, "model_upload_instructions.md")))
        with open(os.path.join(model_dir, "model_upload_instructions.md"), 'r') as read_obj:
            text = read_obj.read()
            self.assertEqual(text[0:20], "/!\\/!\/!\\/!\/!\\   The aggregation model is a special model, please ensure that"[0:20])
        remove_dir_model(model, model_dir)

    def test12_model_aggregation_prepend_line(self):
        '''Test of {{package_name}}.models_training.model_aggregation_regressor.ModelAggregationRegressor.prepend_line'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        model = ModelAggregationRegressor(model_dir=model_dir)
        path = os.path.join(model_dir, 'test.md')
        with open(path, 'w') as f:
            f.write('toto')
        with open(path, 'r') as f:
            self.assertTrue(f.read() == 'toto')
        model.prepend_line(path, 'titi\n')
        with open(path, 'r') as f:
            self.assertTrue(f.read() == 'titi\ntoto')
        remove_dir(model_dir)

    def test13_model_aggregation_regressor_reload_from_standalone(self):
        '''Test of {{package_name}}.models_training.model_aggregation_regressor.ModelAggregationRegressor.reload_from_standalone'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        model_new_dir = os.path.join(os.getcwd(), 'model_new_test_123456789')
        remove_dir(model_dir)

        # Set vars
        x_train = pd.DataFrame({'col_1': [-5, -1, -1, 0, 4.2, 6, 3] * 10, 'col_2': [-8.1, -1.3, -1, 0, 4, 12, 2] * 10})
        x_test = pd.DataFrame({'col_1': [-5], 'col_2': [-8.1]})
        y_train = pd.Series([0, 0, 0, 2, 1, 1, 1] * 10)

        # Create model
        gbt, sgd, _, _ = self.create_models()
        model = ModelAggregationRegressor(model_dir=model_dir, list_models=[gbt, sgd])
        model.fit(x_train, y_train)
        model.save()
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"aggregation_function.pkl")))

        # Reload
        conf_path = os.path.join(model.model_dir, "configurations.json")
        preprocess_pipeline_path = os.path.join(model.model_dir, "preprocess_pipeline.pkl")
        aggregation_function_path = os.path.join(model.model_dir, "aggregation_function.pkl")
        model_new = ModelAggregationRegressor(model_dir=model_new_dir)
        model_new.reload_from_standalone(model_dir=model_dir, configuration_path=conf_path, aggregation_function_path=aggregation_function_path, preprocess_pipeline_path=preprocess_pipeline_path)

        # Test
        self.assertEqual(model.trained, model_new.trained)
        self.assertEqual(model.nb_fit, model_new.nb_fit)
        self.assertEqual(model.x_col, model_new.x_col)
        self.assertEqual(model.y_col, model_new.y_col)
        self.assertEqual(model.level_save, model_new.level_save)
        self.assertEqual([sub_model['name'] for sub_model in model.sub_models], [sub_model['name'] for sub_model in model_new.sub_models])
        self.assertTrue(isinstance(model.sub_models[0]['model'], type(model_new.sub_models[0]['model'])))
        self.assertTrue(isinstance(model.sub_models[1]['model'], type(model_new.sub_models[1]['model'])))
        self.assertEqual(model.aggregation_function.__code__.co_code, model_new.aggregation_function.__code__.co_code)
        preds = model.predict(x_test)
        new_preds = model_new.predict(x_test)
        for pred, new_pred in zip(preds, new_preds):
            self.assertAlmostEqual(pred, new_pred)
        remove_dir_model(model, model_dir)
        remove_dir_model(model_new, model_new_dir)

        ############################################
        # Errors
        ############################################

        model_new = ModelAggregationRegressor(model_dir=model_new_dir)
        with self.assertRaises(FileNotFoundError):
            model_new.reload_from_standalone(model_dir=model_dir, configuration_path='toto.json', aggregation_function_path=aggregation_function_path, preprocess_pipeline_path=preprocess_pipeline_path)
        remove_dir(model_new.model_dir)
        remove_dir(model_new_dir)
        model_new = ModelAggregationRegressor(model_dir=model_new_dir)
        with self.assertRaises(FileNotFoundError):
            model_new.reload_from_standalone(model_dir=model_dir, configuration_path=conf_path, aggregation_function_path='toto.pkl', preprocess_pipeline_path=preprocess_pipeline_path)
        remove_dir(model_new.model_dir)
        remove_dir(model_new_dir)
        model_new = ModelAggregationRegressor(model_dir=model_new_dir)
        with self.assertRaises(FileNotFoundError):
            model_new.reload_from_standalone(model_dir=model_dir, configuration_path=conf_path, aggregation_function_path=aggregation_function_path, preprocess_pipeline_path='toto.pkl')
        remove_dir(model_new.model_dir)
        remove_dir(model_new_dir)

        model_new = ModelAggregationRegressor(model_dir=model_new_dir)
        with self.assertRaises(ValueError):
            model_new.reload_from_standalone(model_dir=model_dir, aggregation_function_path=aggregation_function_path, preprocess_pipeline_path=preprocess_pipeline_path)
        remove_dir(model_new.model_dir)
        remove_dir(model_new_dir)
        model_new = ModelAggregationRegressor(model_dir=model_new_dir)
        with self.assertRaises(ValueError):
            model_new.reload_from_standalone(model_dir=model_dir, configuration_path=conf_path, preprocess_pipeline_path=preprocess_pipeline_path)
        remove_dir(model_new.model_dir)
        remove_dir(model_new_dir)
        model_new = ModelAggregationRegressor(model_dir=model_new_dir)
        with self.assertRaises(ValueError):
            model_new.reload_from_standalone(model_dir=model_dir, configuration_path=conf_path, aggregation_function_path=aggregation_function_path)
        remove_dir(model_new.model_dir)
        remove_dir(model_new_dir)


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()


#!/usr/bin/env python3

## Preprocessing functions
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging
import numpy as np
import pandas as pd
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline, make_pipeline
from sklearn.feature_selection import SelectKBest, SelectorMixin
from sklearn.compose import ColumnTransformer, make_column_selector
from sklearn.feature_extraction.text import CountVectorizer, _VectorizerMixin
from sklearn.preprocessing import FunctionTransformer, StandardScaler, OneHotEncoder

from . import column_preprocessors

# Get logger
logger = logging.getLogger(__name__)


def get_pipelines_dict() -> dict:
    '''Gets a dictionary of available preprocessing pipelines

    Returns:
        dict: Dictionary of preprocessing pipelines
    '''
    pipelines_dict = {
        # - /!\ DO NOT DELETE no_preprocess -> necessary for compatibility /!\ -
        # Identity transformer, hence we specify verbose_feature_names_out to False to not change columns names
        'no_preprocess': ColumnTransformer([('identity', FunctionTransformer(lambda x: x),
                                             make_column_selector())], verbose_feature_names_out=False),
        'preprocess_P1': preprocess_P1(),  # Example of a pipeline
        # 'preprocess_AUTO': preprocess_auto(), # Automatic preprocessing based on statistics on data
        # 'preprocess_P2': preprocess_P2 , ETC ...
    }
    return pipelines_dict


def get_pipeline(pipeline_str: str) -> ColumnTransformer:
    '''Gets a pipeline from its name

    Args:
        pipeline_str (str): Name of the pipeline
    Raises:
        ValueError: If the name of the pipeline is not known
    Returns:
        ColumnTransfomer: Pipeline to be used for the preprocessing
    '''
    # Process
    pipelines_dict = get_pipelines_dict()
    if pipeline_str not in pipelines_dict.keys():
        raise ValueError(f"The pipeline {pipeline_str} is not known.")
    # Get pipeline
    pipeline = pipelines_dict[pipeline_str]
    # Return
    return pipeline


def preprocess_P1() -> ColumnTransformer:
    '''Gets "default" preprocessing pipeline

    Returns:
        ColumnTransformer: The pipeline
    '''
    numeric_pipeline = make_pipeline(SimpleImputer(strategy='median'), StandardScaler())
    # cat_pipeline = make_pipeline(SimpleImputer(strategy='most_frequent'), OneHotEncoder(handle_unknown='ignore'))
    # text_pipeline = make_pipeline(CountVectorizer(), SelectKBest(k=5))

    # Check https://scikit-learn.org/stable/modules/generated/sklearn.compose.make_column_selector.html
    # and https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.select_dtypes.html#pandas.DataFrame.select_dtypes
    # to understand make_column_selector

    # /!\
    # BE VERY CAUTIOUS WHEN USING FunctionTransformer !
    # A pickled pipeline can still depends on a module definition. Hence, even when pickled, you may not have consistent results !
    # Please try to use lambdas or local function, without any dependency. It reduces risk of changes.
    # If you still have to use a module function, try to never change it later on.
    # https://stackoverflow.com/questions/73788824/how-can-i-save-reload-a-functiontransformer-object-and-expect-it-to-always-wor
    # https://github.com/OSS-Pole-Emploi/gabarit/issues/63
    # /!\

    # /!\ EXEMPLE HERE /!\
    # Good practice: Use directly the names of the columns instead of a "selector"
    # WARNING: The text pipeline is supposed to work on a column 'text' -> Please adapt it to your project if you want to use it

    # By default, we only keep the preprocess on numerical columns
    transformers = [
        ('num', numeric_pipeline, make_column_selector(dtype_include='number')),
        # ('cat', cat_pipeline, make_column_selector(dtype_include='category')), # To convert a column in a column with dtype category: df["A"].astype("category")
        # ('text', text_pipeline, 'text'), # CountVectorizer possible one column at a time
    ]

    # TODO: add sparse compatibility !
    # Use somethings like this :
    # - After applying a pipeline ...
    # if scipy.sparse.issparse(preprocessed_x):
    #     preprocessed_df = pd.DataFrame.sparse.from_spmatrix(preprocessed_x)
    # - Before training ...
    # x_train = x_train.sparse.to_coo().tocsr()
    # x_valid = x_valid.sparse.to_coo().tocsr()
    # ...
    pipeline = ColumnTransformer(transformers, sparse_threshold=0, remainder='drop')  # Use remainder='passthrough' to keep all other columns (not recommended)

    return pipeline


# TODO
def preprocess_auto() -> ColumnTransformer:
    '''Gets an "automatic" pipeline. Different functions are applied depending on stats calculated on the data

    Returns:
        ColumnTransformer: The automatic pipeline
    '''
    # Numeric :
    # 1) SimpleImputer()
    # 2) If abs(skew) > 2 && pctl(90) - pctl(10) > 10^3 => logtransform
    # 3) StandardScaler()
    # Categorical :
    # 1) SimpleImputer()
    # 2) If #cat > 5; We accumulate the less represented instances in a meta-category "other"
    # 3) OneHot
    pass


def retrieve_columns_from_pipeline(df: pd.DataFrame, pipeline: ColumnTransformer) -> pd.DataFrame:
    '''Retrieves columns name after preprocessing

    Args:
        df (pd.DataFrame): Dataframe after preprocessing (without target)
        pipeline (ColumnTransformer): Used pipeline
    Raises:
        AttributeError : The pipeline is not fitted
        ValueError : The number of columns is not the same between the pipeline and the preprocessed DataFrame
    Returns:
        pd.DataFrame: Dataframe with columns' name
    '''
    # Use deepcopy !
    new_df = df.copy(deep=True)
    # Check if fitted:
    if not hasattr(pipeline, 'transformers_'):
        raise AttributeError("The pipeline must be fitted to use the function retrieve_columns_from_pipeline")
    # EXPERIMENTAL: We do a try... except... if we can't get the names
    # First try : use sklearn get_feature_names_out function (might crash)
    # Second try : backup on old custom method
    # Third solution : ['x0', 'x1', ...]
    try:
        try:
            new_columns = pipeline.get_feature_names_out()
        # Backup on old custom method
        except:
            new_columns = get_ct_feature_names(pipeline)
        if len(new_columns) != new_df.shape[1]:
            raise ValueError(f"There is a discrepancy in the number of columns between the preprocessed DataFrame ({new_df.shape[1]}) and the pipeline ({len(new_columns)}).")

    # No solution
    except Exception as e:
        logger.error("Can't get the names of the columns. Backup on ['x0', 'x1', ...]")
        logger.error(repr(e))
        new_columns = [f'x{i}' for i in range(len(new_df.columns))]
    # TODO : check for duplicates in new_columns ???
    new_df.columns = new_columns
    return new_df


# Backup solution if get_feature_names_out does not work
# https://scikit-learn.org/stable/auto_examples/release_highlights/plot_release_highlights_1_0_0.html#feature-names-support
def get_ct_feature_names(ct: ColumnTransformer) -> list:
    '''Gets the names of the columns when considering a fitted ColumnTransfomer
    From: https://stackoverflow.com/questions/57528350/can-you-consistently-keep-track-of-column-labels-using-sklearns-transformer-api

    Args:
        ColumnTransformer: Column tranformer to be processed
    Returns:
        list: List of new feature names
    '''
    # Handles all estimators, pipelines inside ColumnTransfomer
    # does not work when remainder =='passthrough'
    # which requires the input column names.
    output_features = []

    for name, estimator, features in ct.transformers_:
        if name != 'remainder':
            if isinstance(estimator, Pipeline):
                current_features = features
                for step in estimator:
                    if type(step) == tuple:
                        step = step[1]
                    current_features = get_feature_out(step, current_features)
                features_out = current_features
            else:
                features_out = get_feature_out(estimator, features)
            if hasattr(ct, 'verbose_feature_names_out') and ct.verbose_feature_names_out == False:
                output_features.extend(features_out)
            else:
                output_features.extend([f'{name}__{feat}' for feat in features_out])
        elif estimator == 'passthrough':
            # features is indexes in case of passthrough
            if hasattr(ct, 'verbose_feature_names_out') and ct.verbose_feature_names_out == False:
                output_features.extend(ct.feature_names_in_[features])
            else:
                output_features.extend([f'remainder__{feat}' for feat in ct.feature_names_in_[features]])

    return output_features


# Backup solution if get_feature_names_out does not work
def get_feature_out(estimator, features_in: list) -> list:
    '''Gets the name of a column when considering a fitted estimator
    From: https://stackoverflow.com/questions/57528350/can-you-consistently-keep-track-of-column-labels-using-sklearns-transformer-api

    Args:
        (?): Estimator to be processed
        (list): Input columns
    Returns:
        list: List of new feature names
    '''
    if hasattr(estimator, 'get_feature_names'):
        if isinstance(estimator, _VectorizerMixin):
            # handling all vectorizers
            return estimator.get_feature_names()
        else:
            return estimator.get_feature_names(features_in)
    elif isinstance(estimator, SelectorMixin):
        return np.array(features_in)[estimator.get_support()]
    else:
        return features_in


if __name__ == '__main__':
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3

## Definition of a class to abstract how MlFlow works
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Classes :
# - MLflowLogger -> Abstracts how MlFlow works


import os
import math
import mlflow
import pathlib
import logging
import pandas as pd
from typing import Union
from urllib.parse import urlparse
from matplotlib.figure import Figure

from .. import utils

# Deactivation of GIT warning for mlflow
os.environ["GIT_PYTHON_REFRESH"] = "quiet"

class MLflowLogger:
    '''Abstracts how MlFlow works'''

    def __init__(self, experiment_name: str, tracking_uri: str = '', artifact_uri: str = '') -> None:
        '''Class initialization
        Args:
            experiment_name (str):  Name of the experiment to activate
        Kwargs:
            tracking_uri (str): URI of the tracking server
            artifact_uri (str): URI where to store artifacts
        '''
        # Get logger
        self.logger = logging.getLogger(__name__)

        # Backup to local save if no uri (i.e. empty string)
        if not tracking_uri:
            tracking_uri = pathlib.Path(os.path.join(utils.get_data_path(), 'experiments', 'mlruns')).as_uri()
        # Add "file" scheme if no scheme in the tracking_uri
        elif not urlparse(tracking_uri).scheme:
            tracking_uri = pathlib.Path(tracking_uri).resolve().as_uri()

        # If no artifact_uri and tracking_uri scheme is "file", we set a default artifact_uri in experiments folder
        # Otherwise we suppose artifact_uri is configured by the system
        if not artifact_uri and urlparse(tracking_uri).scheme == "file":
            artifact_uri = pathlib.Path(os.path.join(utils.get_data_path(), 'experiments', 'mlruns_artifacts')).as_uri()

        # Set tracking URI & experiment name
        self.tracking_uri = tracking_uri
        
        # Get the experiment if it exists and check if there is a connection error by doing it
        try:
            experiment = mlflow.get_experiment_by_name(experiment_name)
        except Exception as e:
            self.logger.error(repr(e))
            raise ConnectionError(f"Can't reach MLflow at {self.tracking_uri}. Please check the URI.")

        # If the experiment exists, we recover experiment id and artifact_uri (which is link to the experiment)
        if experiment:
            experiment_id = experiment.experiment_id
            artifact_uri = experiment.artifact_location
        # Otherwise we create a new experiment with the provided artifact_uri
        else:
            experiment_id = mlflow.create_experiment(experiment_name, artifact_location=artifact_uri)
            experiment = mlflow.get_experiment_by_name(experiment_name)
            artifact_uri = experiment.artifact_location
        
        mlflow.set_experiment(experiment_id=experiment_id)

        self.__experiment_id = experiment_id
        self.__experiment_name = experiment_name
        self.__artifact_uri = artifact_uri

        self.logger.info(f'MLflow running. Metrics available @ {self.tracking_uri}. Experiment artifacts availaible @ {self.artifact_uri}')

    @property
    def tracking_uri(self) -> str:
        '''Current tracking uri'''
        return mlflow.get_tracking_uri()

    @tracking_uri.setter
    def tracking_uri(self, uri:str) -> None:
        '''Set tracking uri'''
        mlflow.set_tracking_uri(uri)

    @property
    def experiment_id(self) -> str:
        '''Experiment id. It can not be changed.'''
        return self.__experiment_id

    @property
    def experiment_name(self) -> str:
        '''Experiment name. It can not be changed.'''
        return self.__experiment_name

    @property
    def artifact_uri(self) -> str:
        '''Experiment artifact URI. It can not be changed.'''
        return self.__artifact_uri

    def end_run(self) -> None:
        '''Stops an MLflow run'''
        try:
            mlflow.end_run()
        except Exception:
            self.logger.error("Can't stop mlflow run")

    def log_metric(self, key: str, value, step: Union[int, None] = None) -> None:
        '''Logs a metric on mlflow

        Args:
            key (str): Name of the metric
            value (float, ?): Value of the metric
        Kwargs:
            step (int): Step of the metric
        '''
        # Check for None
        if value is None:
            value = math.nan
        # Log metric
        mlflow.log_metric(key, value, step)

    def log_metrics(self, metrics: dict, step: Union[int, None] = None) -> None:
        '''Logs a set of metrics in mlflow

        Args:
            metrics (dict): Metrics to log
        Kwargs:
            step (int): Step of the metric
        '''
        # Check for Nones
        for k, v in metrics.items():
            if v is None:
                metrics[k] = math.nan
        # Log metrics
        mlflow.log_metrics(metrics, step)

    def log_param(self, key: str, value) -> None:
        '''Logs a parameter in mlflow

        Args:
            key (str): Name of the parameter
            value (str, ?): Value of the parameter (which will be cast to str if not already of type str)
        '''
        if value is None:
            value = 'None'
        # Log parameter
        mlflow.log_param(key, value)

    def log_params(self, params: dict) -> None:
        '''Logs a set of parameters in mlflow

        Args:
            params (dict): Name and value of each parameter
        '''
        # Check for Nones
        for k, v in params.items():
            if v is None:
                params[k] = 'None'
        # Log parameters
        mlflow.log_params(params)

    def set_tag(self, key: str, value) -> None:
        '''Logs a tag in mlflow

        Args:
            key (str): Name of the tag
            value (str, ?): Value of the tag (which will be cast to str if not already of type str)
        Raises:
            ValueError: If the object value is None
        '''
        if value is None:
            raise ValueError('value must not be None')
        # Log tag
        mlflow.set_tag(key, value)

    def set_tags(self, tags: dict) -> None:
        '''Logs a set of tags in mlflow

        Args:
            tags (dict): Name and value of each tag
        '''
        # Log tags
        mlflow.set_tags(tags)

    def valid_name(self, key: str) -> bool:
        '''Validates key names

        Args:
            key (str): Key to check
        Returns:
            bool: If key is a valid mlflow key
        '''
        if mlflow.mlflow.utils.validation._VALID_PARAM_AND_METRIC_NAMES.match(key):
            return True
        else:
            return False

    def log_df_stats(self, df_stats: pd.DataFrame, label_col: str = 'Label') -> None:
        '''Log a dataframe containing metrics from a training

        Args:
            df_stats (pd.Dataframe): Dataframe containing metrics from a training
        Kwargs:
            label_col (str): default labelc column name
        '''
        if label_col not in df_stats.columns:
            raise ValueError(f"The provided label column name ({label_col}) not found in df_stats' columns.")

        # Get metrics columns
        metrics_columns = [col for col in df_stats.columns if col != label_col]

        # Log labels
        labels = df_stats[label_col].values
        for i, label in enumerate(labels):  # type: ignore
            self.log_param(f'Label {i}', label)

        # Log metrics
        ml_flow_metrics = {}
        for i, row in df_stats.iterrows():
            for j, col in enumerate(metrics_columns):
                metric_key = f"{row[label_col]} --- {col}"
                # Check that mlflow accepts the key, otherwise, replace it
                # TODO: could be improved ...
                if not self.valid_name(metric_key):
                    metric_key = f"Label {i} --- {col}"
                if not self.valid_name(metric_key):
                    metric_key = f"{row[label_col]} --- Col {j}"
                if not self.valid_name(metric_key):
                    metric_key = f"Label {i} --- Col {j}"
                ml_flow_metrics[metric_key] = row[col]

        # Log metrics
        self.log_metrics(ml_flow_metrics)

    def log_dict(self, dictionary: dict, artifact_file: str) -> None:
        '''Logs a dictionary as an artifact in MLflow

        Args:
            dictionary (dict): A dictionary
            artifact_file (str): The run-relative artifact file path in posixpath format to which the dictionary is saved
        '''
        mlflow.log_dict(dictionary=dictionary, artifact_file=artifact_file)

    def log_text(self, text: str, artifact_file: str) -> None:
        '''Logs a text as an artifact in MLflow

        Args:
            text (str): A text
            artifact_file (str): The run-relative artifact file path in posixpath format to which the dictionary is saved
        '''
        mlflow.log_text(text=text, artifact_file=artifact_file)

    def log_figure(self, figure: Figure, artifact_file: str) -> None:
        '''Logs a text as an artifact in MLflow

        Args:
            figure (matplotlib.figure.Figure): A matplotlib figure
            artifact_file (str): The run-relative artifact file path in posixpath format to which the figure is saved
        '''
        mlflow.log_figure(figure=figure, artifact_file=artifact_file)


if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import pytest
from fastapi.testclient import TestClient
from .create_test_model import TestExplainer
from {{package_name}}.core.resources import RESOURCES, RESOURCE_MODEL


def test_predict(test_complete_client: TestClient):
    """Test the route predict thanks to the TestModel we created in conftest.py"""
    response = test_complete_client.post(
        "/tests/predict", json={"content": ["gab", "gabarit"]}
    )
    assert response.status_code == 200
    assert response.json() == [
        {"probability": pytest.approx(3 / 7)},
        {"probability": 1},
    ]

def test_explain(test_complete_client: TestClient):
    """Test the route explain"""
    # 501 HTML error
    response = test_complete_client.post(
        "/tests/explain",
        json={"content": ["gab", "gabarit"]},
    )
    assert response.status_code == 501

    # 501 JSON error
    response = test_complete_client.post(
        "/tests/explain",
        json={"content": ["gab", "gabarit"]},
        headers={"Accept": "application/json"},
    )
    assert response.status_code == 501

    # Now add a TestExplainer to test the explanations
    RESOURCES[RESOURCE_MODEL]._model_explainer = TestExplainer()

    # HTML response
    response = test_complete_client.post(
        "/tests/explain",
        json={"content": ["gab", "gabarit"]},
    )
    assert response.status_code == 200
    assert "<li>gab : GAB____</li>" in response.text

    # JSON response
    response = test_complete_client.post(
        "/tests/explain",
        json={"content": ["gab", "gabarit"]},
        headers={"Accept": "application/json"},
    )
    assert response.status_code == 200
    assert response.json() == [
        {"common_letters": "GAB____"},
        {"common_letters": "GABARIT"},
    ]



#!/usr/bin/env python3

## Training a model - Classification task
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Ex: python 2_training.py  --filename dataset_train_preprocess_P1.csv --filename_valid dataset_valid_preprocess_P1.csv --x_col preprocessed_text --y_col Survived


import os
# Disable some tensorflow logs right away
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import gc
import time
import logging
import argparse
import warnings
import numpy as np
import pandas as pd
from functools import partialmethod
from typing import Union, List, Type, Tuple

from {{package_name}} import utils
from {{package_name}}.preprocessing import preprocess
from {{package_name}}.monitoring.mlflow_logger import MLflowLogger
from {{package_name}}.models_training.model_class import ModelClass
from {{package_name}}.models_training import (
    utils_models,
    model_aggregation,
    model_huggingface,
)
from {{package_name}}.models_training.models_sklearn import (
    model_tfidf_gbt,
    model_tfidf_lgbm,
    model_tfidf_sgdc,
    model_tfidf_svm,
)
from {{package_name}}.models_training.models_tensorflow import (
    model_embedding_cnn,
    model_embedding_lstm,
    model_embedding_lstm_attention,
    model_embedding_lstm_structured_attention,
    model_embedding_lstm_gru,
    model_tfidf_dense,
)

# Disable some warnings
warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)

# Get logger
logger = logging.getLogger('{{package_name}}.2_training')


def main(filename: str, x_col: Union[str, int], y_col: List[Union[str, int]], filename_valid: Union[str, None] = None,
         min_rows: Union[int, None] = None, level_save: str = 'HIGH',
         sep: str = '{{default_sep}}', encoding: str = '{{default_encoding}}',
         model: Union[Type[ModelClass], None] = None,
         mlflow_experiment: Union[str, None] = None) -> None:
    '''Trains a model

    Args:
        filename (str): Name of the training dataset (actually a path relative to {{package_name}}-data)
        x_col (str | int): Name of the model's input column - x
        y_col (list<str|int>): Name of the model's target column(s) - y
    Kwargs:
        min_rows (int): Minimal number of occurrences for a class to be considered by the model
            Corresponding entries are removed from both training & validation dataset - mono-label only
        filename_valid (str): Name of the validation dataset (actually a path relative to {{package_name}}-data)
            If None, we do not use a validation dataset.
                -> for keras models (i.e. Neural Networks), we'll use a portion of the training dataset as the validation.
        level_save (str): Save level
            LOW: statistics + configurations + logger keras - /!\\ the model won't be reusable /!\\ -
            MEDIUM: LOW + hdf5 + pkl + plots
            HIGH: MEDIUM + predictions
        sep (str): Separator to use with the .csv files
        encoding (str): Encoding to use with the .csv files
        model (ModelClass): A model to be fitted. This should only be used for testing purposes.
        mlflow_experiment (str): Name of the current experiment. If None, no experiment will be saved.
    Raises:
        ValueError: If level_save value is not a valid option (['LOW', 'MEDIUM', 'HIGH'])
        ValueError: If multi-labels and bad OHE format
    '''
    logger.info("Training a model ...")

    if level_save not in ['LOW', 'MEDIUM', 'HIGH']:
        raise ValueError(f"level_save value ({level_save}) is not a valid option (['LOW', 'MEDIUM', 'HIGH']).")


    ##############################################
    # Manage training dataset
    ##############################################

    # Get dataset
    df_train, preprocess_str = load_dataset(filename, sep=sep, encoding=encoding)

    ### INFO
    ### EACH MODEL needs the same target data format:
    ###
    ###   - OHE (integers) if multi-label
    ###     e.g.
    ###         col1 col2 col3
    ###            1    0   1
    ###            0    0   0
    ###            0    1   0
    ###
    ###   - a single string column if mono-label (even if 0/1 -> '0'/'1')
    ###     e.g.
    ###         target
    ###           target_1
    ###           target_2
    ###           target_1
    ###           target_3
    ###
    ### Below are some examples to preprocess your datasets with the correct format:
    ###
    ###
    ### - multi-label -> Transform tuples in OHE format
    ### ********
    ### y_col = .... # Column with target in tuple format
    ### # Make sure column is tuples and not strings (optional)
    ### from ast import literal_eval
    ### df_train[y_col] = df_train[y_col].apply(lambda x: literal_eval(x))
    ### # Transform to OHE
    ### df_train, y_col = utils_models.preprocess_model_multilabel(df_train, y_col, classes=None)
    ### ********

    # Check if multi-label, i.e. several target columns
    if len(y_col) > 1:
        multi_label = True
        df_train[y_col] = df_train[y_col].astype(int)  # Need to cast OHE var into integers
        col_not_ohe = []
        for col in y_col:
            if sorted(df_train[col].unique()) != [0, 1]:
                col_not_ohe.append(col)
        if len(col_not_ohe):
            raise ValueError(f"You provided several target columns, but the columns {col_not_ohe} seem not to be in the correct OHE format.")
            
    else:
        multi_label = False
        y_col = y_col[0]
        # No need to cast target in string, already done by the data loader


    ##############################################
    # Manage validation dataset
    ##############################################

    # Get valid dataset (/!\ we consider that this dataset has the same preprocessing as the training set /!\)
    if filename_valid is not None:
        logger.info(f"Using file {filename_valid} as our validation set.")
        df_valid, preprocess_str_valid = load_dataset(filename_valid, sep=sep, encoding=encoding)
        if preprocess_str_valid != preprocess_str:
            logger.warning("Validation set and training set does not expose the same preprocessing metadata.")
            logger.warning(f"Train : {preprocess_str}")
            logger.warning(f"Valid : {preprocess_str_valid}")
            logger.warning("That will probably lead to bad results !")
            logger.warning("Still continuing...")
        ### INFO: the validation set must have a correct format (cf. traing set info above)
        # Manage OHE format (ensure int)
        if multi_label:
            df_valid[y_col] = df_valid[y_col].astype(int)
        # No need to cast in str if not multilabel, already done by the data loader
    else:
        logger.info("No validation set provided.")
        logger.info("In case of Keras models, we'll use a portion of the training dataset as the validation")


    ##############################################
    # Manage classes
    ##############################################

    # Remove small classes if wanted (only possible if not multi-labels)
    if min_rows is not None and not multi_label:
        df_train = utils_models.remove_small_classes(df_train, y_col, min_rows=min_rows)

    # Remove classes in valid, but not in train (only possible if not multi-labels)
    # If we do not do that, some model (e.g. Keras models) might not work as intended
    # To get metrics on whole dataset, one can use 4_predict.py
    if filename_valid is not None and not multi_label:
        train_classes = list(df_train[y_col].unique())
        valid_classes = list(df_valid[y_col].unique())
        for cl in valid_classes:
            if cl not in train_classes:
                logger.warning(f"Removing class {cl} from the valid dataset (not present in the training dataset).")
        df_valid = df_valid[df_valid[y_col].isin(train_classes)].reset_index(drop=True)


    ##############################################
    # Manage input data
    ##############################################

    # We do not check for columns presence in dataframes
    # TODO: add checks ?
    # Get x, y for both train and valid
    x_train, y_train = df_train[x_col], df_train[y_col]
    if filename_valid is not None:
        x_valid, y_valid = df_valid[x_col], df_valid[y_col]
    else:
        x_valid, y_valid = None, None


    ##############################################
    # Model selection
    ##############################################

    # INFO
    # If you want to continue training of a model, it needs to be reloaded here (only some models are compatible)
    # model, _ = utils_models.load_model("dir_model")
    # Then, it is possible to change some parameters such as the learning rate
    # Be careful to work with the same preprocessing as the first training

    # INFO
    # To use hyperparameters tuning, we can do something like this (experimental, needs improvements):
    # model_cls = model_tfidf_svm.ModelTfidfSvm  # Model Class to be tuned
    # model_params = {'x_col': x_col, 'y_col': y_col, 'multi_label': multi_label}  # Model's fixed parameters
    # hp_params = {'tfidf_params': [{'analyzer': 'word', 'ngram_range': (1, 2), "max_df":0.1}, {'analyzer': 'word', 'ngram_range': (1, 3), "max_df":0.1}]}  # Parameters to be tested
    # scoring_fn = "f1"  # Scoring function to MAXIMIZE
    # kwargs_fit = {'x_train': x_train, 'y_train': y_train, 'with_shuffle': True}  # Fit arguments (i.e. training data)
    # n_splits = 5  # Number of crossvalidation
    # model = utils_models.search_hp_cv(model_cls, model_params, hp_params, scoring_fn, kwargs_fit, n_splits=n_splits)  # Returns a model with the "best" parameters, to be fitted on the whole dataset


    if model is None:
        model = model_tfidf_svm.ModelTfidfSvm(x_col=x_col, y_col=y_col, level_save=level_save,
                                              tfidf_params={'analyzer': 'word', 'ngram_range': (1, 2), 'min_df': 1, 'max_df': 0.25, 'max_features': 100000},
                                              svc_params={'C': 1.0, 'max_iter': 10000},
                                              multi_label=multi_label)
        # model = model_tfidf_gbt.ModelTfidfGbt(x_col=x_col, y_col=y_col, level_save=level_save,
        #                                       tfidf_params={'analyzer': 'word', 'ngram_range': (1, 2), 'min_df': 1, 'max_df': 0.25, 'max_features': 100000},
        #                                       gbt_params={'learning_rate': 0.1, 'n_estimators': 100, 'max_depth': 10, 'subsample': 1.0, 'max_features': 'auto'},
        #                                       multi_label=multi_label)
        # model = model_tfidf_lgbm.ModelTfidfLgbm(x_col=x_col, y_col=y_col, level_save=level_save,
        #                                         tfidf_params={'analyzer': 'word', 'ngram_range': (1, 2), 'min_df': 1, 'max_df': 0.25, 'max_features': 100000},
        #                                         lgbm_params={'learning_rate': 0.1, 'n_estimators': 100, 'max_depth': 5, 'subsample': 1.0, 'num_leaves': 127},
        #                                         multi_label=multi_label)
        # model = model_tfidf_sgdc.ModelTfidfSgdc(x_col=x_col, y_col=y_col, level_save=level_save,
        #                                         tfidf_params={'analyzer': 'word', 'ngram_range': (1, 2), 'min_df': 1, 'max_df': 0.25, 'max_features': 100000},
        #                                         sgdc_params={'loss': 'hinge', 'max_iter': 1000},
        #                                         multi_label=multi_label)
        # model = model_embedding_lstm.ModelEmbeddingLstm(x_col=x_col, y_col=y_col, level_save=level_save,
        #                                                 batch_size=64, epochs=99, patience=5,
        #                                                 max_sequence_length=200, max_words=100000,
        #                                                 multi_label=multi_label)
        # model = model_embedding_lstm_attention.ModelEmbeddingLstmAttention(x_col=x_col, y_col=y_col, level_save=level_save,
        #                                                                    batch_size=64, epochs=99, patience=5,
        #                                                                    max_sequence_length=200, max_words=100000,
        #                                                                    multi_label=multi_label)
        # model = model_embedding_lstm_structured_attention.ModelEmbeddingLstmStructuredAttention(x_col=x_col, y_col=y_col, level_save=level_save,
        #                                                                                         batch_size=64, epochs=99, patience=5,
        #                                                                                         max_sequence_length=200, max_words=100000,
        #                                                                                         multi_label=multi_label)
        # model = model_embedding_lstm_gru.ModelEmbeddingLstmGru(x_col=x_col, y_col=y_col, level_save=level_save,
        #                                                               batch_size=64, epochs=99, patience=5,
        #                                                               max_sequence_length=60, max_words=100000,
        #                                                               multi_label=multi_label)
        # model = model_embedding_cnn.ModelEmbeddingCnn(x_col=x_col, y_col=y_col, level_save=level_save,
        #                                               batch_size=64, epochs=99, patience=5,
        #                                               max_sequence_length=200, max_words=100000,
        #                                               multi_label=multi_label)
        # model = model_tfidf_dense.ModelTfidfDense(x_col=x_col, y_col=y_col, level_save=level_save,
        #                                           batch_size=64, epochs=99, patience=5,
        #                                           tfidf_params={'analyzer': 'word', 'ngram_range': (1, 2), 'min_df': 1, 'max_df': 0.25, 'max_features': 100000},
        #                                           multi_label=multi_label)
        # model = model_huggingface.ModelHuggingFace(x_col=x_col, y_col=y_col, level_save=level_save,
        #                                            batch_size=64, epochs=99, patience=5,
        #                                            transformer_name='Geotrend/distilbert-base-fr-cased',
        #                                            multi_label=multi_label)
        # model = model_aggregation.ModelAggregation(x_col=x_col, y_col=y_col, level_save=level_save,
        #                                            list_models=[model_tfidf_svm.ModelTfidfSvm(), model_tfidf_svm.ModelTfidfSvm()],
        #                                            multi_label=multi_label)

    # Display if GPU is being used
    model.display_if_gpu_activated()


    ##############################################
    # Train the model !
    ##############################################

    start_time = time.time()
    logger.info("Starting training the model ...")
    model.fit(x_train, y_train, x_valid=x_valid, y_valid=y_valid, with_shuffle=True)
    fit_time = time.time() - start_time


    ##############################################
    # Save trained model
    ##############################################

    # Save model
    model.save(
        json_data={
            'filename': filename,
            'filename_valid': filename_valid,
            'min_rows': min_rows,
            'preprocess_str': preprocess_str,
            'fit_time': f"{round(fit_time, 2)}s",
        }
    )
    logger.info(f"Model {model.model_name} saved in directory {model.model_dir}")


    ##############################################
    # Model metrics
    ##############################################

    # Series to add
    cols_to_add: List[pd.Series] = []  # You can add columns to save here
    series_to_add_train = [df_train[col] for col in cols_to_add]
    series_to_add_valid = [df_valid[col] for col in cols_to_add]
    gc.collect()  # In some cases, helps with OOMs

    # Get results
    y_pred_train = model.predict(x_train, return_proba=False)
    df_stats = model.get_and_save_metrics(y_train, y_pred_train, x=x_train, series_to_add=series_to_add_train, type_data='train')
    gc.collect()  # In some cases, helps with OOMs
    # Get predictions on valid if exists
    if x_valid is not None:
        y_pred_valid = model.predict(x_valid, return_proba=False)
        df_stats = model.get_and_save_metrics(y_valid, y_pred_valid, x=x_valid, series_to_add=series_to_add_valid, type_data='valid')
        gc.collect()  # In some cases, helps with OOMs


    ##############################################
    # Logger MLflow
    ##############################################

    # Logging metrics on MLflow
    if mlflow_experiment:
        # Get logger
        mlflow_logger = MLflowLogger(
            experiment_name=f"{{package_name}}/{mlflow_experiment}",
            tracking_uri="{{mlflow_tracking_uri}}",
            artifact_uri="{{mlflow_artifact_uri}}",
        )
        # Set model name, save metrics & configurations
        mlflow_logger.set_tag('model_name', f"{os.path.basename(model.model_dir)}")
        mlflow_logger.log_df_stats(df_stats)
        mlflow_logger.log_dict(model.json_dict, "configurations.json")
        # To log more tags/params, you can use mlflow_logger.set_tag(key, value) or mlflow_logger.log_param(key, value)
        # Log a sweetviz report
        report = get_sweetviz_report(df_train=df_train, y_pred_train=y_pred_train, y_col=y_col,
                                     df_valid=df_valid if filename_valid else None,
                                     y_pred_valid=y_pred_valid if filename_valid else None)
        if report:
            mlflow_logger.log_text(report, "sweetviz_train_valid.html")
        # Stop MLflow if started
        mlflow_logger.end_run()


def load_dataset(filename: str, sep: str = '{{default_sep}}', encoding: str = '{{default_encoding}}') -> Tuple[pd.DataFrame, str]:
    '''Function to load a dataset - retrieves preprocessing metadata

    Args:
        filename (str): Name of the dataset to load (actually a path relative to {{package_name}}-data)
    Kwargs:
        sep (str): Separator to use with the .csv files
        encoding (str): Encoding to use with the .csv files
    Raises:
        FileNotFoundError: If the file does not exist in {{package_name}}-data
    Returns:
        pd.DataFrame: Loaded dataframe
        str: Preprocessing used on this dataset (from metadata)
    '''
    logger.info(f"Loading a dataset ({filename})")

    # Get dataset
    data_path = utils.get_data_path()
    file_path = os.path.join(data_path, filename)
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"The file {file_path} does not exist")

    # Load dataset
    df, first_line = utils.read_csv(file_path, sep=sep, encoding=encoding, dtype=str)
    df = df.fillna('')  # Compulsory as some sentences might be empty after preprocessing, and reloaded as NAs

    # Get preprocess type
    if first_line is not None and first_line.startswith('#'):
        preprocess_str = first_line[1:]  # remove # (sharp)
    else:
        preprocess_str = 'no_preprocess'

    # Return
    return df, preprocess_str


def get_sweetviz_report(df_train: pd.DataFrame, y_pred_train: np.ndarray, y_col: Union[str, List[str]],
                        df_valid: Union[pd.DataFrame, None], y_pred_valid: Union[np.ndarray, None]) -> str:
    '''Generate a sweetviz report that can be logged into MLflow

    Args:
        df_train (pd.DataFrame): Training data
        y_pred_train (np.ndarray): Model predictions on training data
        y_col (str | list<str>): Target(s) column(s)
        df_valid (pd.DataFrame): Validation data
        y_pred_valid (np.ndarray): Model predictions on validation data
    Returns:
        str: A HTML sweetviz report
    '''
    logger.info("Producing a sweetviz report ...")

    # SweetViz add too much logs to be use in production
    # https://github.com/fbdesignpro/sweetviz/issues/124
    # Deactivate tqdm and import sweetviz
    try:
        from tqdm import tqdm
        tqdm.__init__ = partialmethod(tqdm.__init__, disable=True)
    except ImportError:
        pass

    try:
        import sweetviz
        import tempfile
    except ImportError:
        return None

    # Add predictions to our datasets
    # First, get new columns name
    y_col_names = [f"pred_{col}" for col in y_col] if isinstance(y_col, list) else [f"pred_{y_col}"]
    # Add predictions to train data
    df_train.loc[:, y_col_names] = y_pred_train
    # Add predictions to validation data
    if df_valid is not None and y_pred_valid is not None:
        df_valid.loc[:, y_col_names] = y_pred_valid

    # Try to specify target feature. That could fail due to the fact that sweetviz
    # can not handle multiple target columns and all possible target types.
    if isinstance(y_col, str):
        target = y_col
        # IF only 0 & 1 in string, cast target column to int
        if sorted(df_train[target].unique()) == ['0', '1']:
            df_train[target] = df_train[target].astype(int)
            if df_valid is not None:
                df_valid[target] = df_valid[target].astype(int)
    else:
        target = None
    # Prepare SweetViz datasets
    train_data = [df_train, "Training data"]
    valid_data = [df_valid, "Validation data"] if df_valid is not None else None

    # Get sweetviz report
    # Strategy : if target not None, try to get report with target
    # If it fails, or if target is None, backup without target
    # pairwise_analysis must be set to off, otherwise takes far too much time
    get_report_without_target = True
    report: sweetviz.DataframeReport
    if target is not None:
        try:
            report = sweetviz.compare(train_data, valid_data, target_feat=target, pairwise_analysis="off")
            get_report_without_target = False  # No need to get report without target
        except (KeyError, ValueError):
            logger.info("Can't produce a sweetviz report with 'target_feat'. Proceeding without this option.")
    if get_report_without_target:
        report = sweetviz.compare(train_data, valid_data, pairwise_analysis="off")

    # We need to call show_html in order to get _page_html
    # Hence, we'll do it on a temp file
    with tempfile.TemporaryDirectory() as tmp_dirname:
        tmp_file = os.path.join(tmp_dirname, "report.html")
        report.show_html(tmp_file, open_browser=False, layout="vertical")

    # Return html code as string
    return report._page_html


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--filename', default='dataset_preprocess_P1.csv', help="Name of the training dataset (actually a path relative to {{package_name}}-data)")
    parser.add_argument('-x', '--x_col', default='preprocessed_text', help="Name of the model's input column - x")
    parser.add_argument('-y', '--y_col', nargs='+', required=True, help="Name of the model's target column(s) - y")
    parser.add_argument('-m', '--min_rows', type=int, default=None, help="Minimal number of occurrences for a class to be considered by the model")
    parser.add_argument('--filename_valid', default=None, help="Name of the validation dataset (actually a path relative to {{package_name}}-data)")
    parser.add_argument('-l', '--level_save', default='HIGH', help="Save level -> ['LOW', 'MEDIUM', 'HIGH']")
    parser.add_argument('--sep', default='{{default_sep}}', help="Separator to use with the .csv files.")
    parser.add_argument('--encoding', default="{{default_encoding}}", help="Encoding to use with the .csv files.")
    parser.add_argument('--force_cpu', dest='on_cpu', action='store_true', help="Whether to force training on CPU (and not GPU)")
    parser.add_argument('--mlflow_experiment', help="Name of the current experiment. MLflow tracking is activated only if fulfilled.")
    parser.set_defaults(on_cpu=False)
    args = parser.parse_args()
    # Check forced CPU usage
    if args.on_cpu:
        os.environ['CUDA_VISIBLE_DEVICES'] = ""
        logger.info("----------------------------")
        logger.info("CPU USAGE FORCED BY THE USER")
        logger.info("----------------------------")
    # Main
    main(filename=args.filename, x_col=args.x_col, y_col=args.y_col,
         min_rows=args.min_rows, filename_valid=args.filename_valid,
         level_save=args.level_save, sep=args.sep, encoding=args.encoding,
         mlflow_experiment=args.mlflow_experiment)
